import sys
import datetime
import platform
import getpass
import os
import requests
import subprocess
import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)
from functools import wraps
from triage.util.db import scoped_session, get_for_update
from triage.util.introspection import classpath
from triage import __version__

try:
    try:
        from pip._internal.operations import freeze as pip_freeze
    except ImportError:  # pip < 10.0
        from pip.operations import freeze as pip_freeze
except ImportError:
    pip_freeze = None


from triage.component.results_schema import ExperimentRun, ExperimentRunStatus


def infer_git_hash():
    """Attempt to infer the git hash of the repository in the current working directory

    Returns: Either the 'git rev-parse HEAD' output or None
    """
    try:
        git_hash = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode('utf-8')
    except Exception as exc:
        logger.spam("Unable to infer git hash")
        git_hash = None
    return git_hash


def infer_triage_version():
    return __version__

def infer_python_version():
    """ Returns python version """
    return sys.version.replace("\r", "").replace("\n", "")

def infer_installed_libraries():
    """Attempt to infer the installed libraries by running pip freeze and formatting as a list

    Returns: Either a list, or None
    """
    if pip_freeze is not None:
        installed_libraries = pip_freeze.freeze()
    else:
        logger.spam("Unable to pip freeze, cannot list installed libraries")
        installed_libraries = []
    return installed_libraries


def infer_ec2_instance_type():
    """Attempt to infer the instance type of the ec2 instance by querying Amazon's endpoint

    Returns: Either the ec2 instance type as returned by Amazon's endpoint, or None
    """
    try:
        ec2_instance_type = requests.get(
            'http://169.254.169.254/latest/meta-data/instance-type',
            timeout=0.01
        ).text
    except requests.exceptions.RequestException:
        logger.spam(
            "Unable to retrieve metadata about ec2 instance, will not set ec2 instance type"
        )
        ec2_instance_type = None
    return ec2_instance_type


def infer_log_location():
    """Attempt to infer the location of the log file of the root logger

    Returns: Either the baseFilename of the first FileHandler on the root logger, or None
    """
    root_logger_handlers = [
        handler
        for handler in logging.getLoggerClass().root.handlers
        if isinstance(handler, logging.FileHandler)
    ]
    if root_logger_handlers:
        log_location = root_logger_handlers[0].baseFilename
    else:
        logger.spam("No FileHandler found in root logger, cannot record logging filename")
        log_location = None
    return log_location


def initialize_tracking_and_get_run_id(
    experiment_hash,
    experiment_class_path,
    random_seed,
    experiment_kwargs,
    db_engine
):
    """Create a row in the ExperimentRun table with some initial info and return the created run_id

    Args:
        experiment_hash (str) An experiment hash that exists in the experiments table
        experiment_class_path (str) The name of the experiment subclass used
        random_seed (int) Random seed used to run the experiment
        experiment_kwargs (dict) Any runtime Experiment keyword arguments that should be saved
        db_engine (sqlalchemy.engine)
    """
    # Any experiment kwargs that are types (e.g. MatrixStorageClass) can't
    # be serialized, so just use the class name if so
    cleaned_experiment_kwargs = {
        k: (classpath(v) if isinstance(v, type) else v)
        for k, v in experiment_kwargs.items()
    }
    run = ExperimentRun(
        start_time=datetime.datetime.now(),
        git_hash=infer_git_hash(),
        triage_version=infer_triage_version(),
        python_version=infer_python_version(),
        experiment_hash=experiment_hash,
        last_updated_time=datetime.datetime.now(),
        current_status=ExperimentRunStatus.started,
        installed_libraries=infer_installed_libraries(),
        platform=platform.platform(),
        os_user=getpass.getuser(),
        working_directory=os.getcwd(),
        ec2_instance_type=infer_ec2_instance_type(),
        log_location=infer_log_location(),
        experiment_class_path=experiment_class_path,
        random_seed = random_seed,
        experiment_kwargs=cleaned_experiment_kwargs,
    )
    run_id = None
    with scoped_session(db_engine) as session:
        session.add(run)
        session.commit()
        run_id = run.run_id
    if not run_id:
        raise ValueError("Failed to retrieve run_id from saved row")
    return run_id


def get_run_for_update(db_engine, run_id):
    """Yields an ExperimentRun at the given run_id for update

    Will kick the last_update_time timestamp of the row each time.

    Args:
        db_engine (sqlalchemy.engine)
        run_id (int) The identifier/primary key of the run
    """
    return get_for_update(db_engine, ExperimentRun, run_id)


def experiment_entrypoint(entrypoint_func):
    """Decorator to control tracking of an experiment run at the wrapped method

    To update the database, it requires the instance of the wrapped method to have a
    db_engine and run_id.

    Upon method entry, will update the ExperimentRun row with the wrapped method name.
    Upon method exit, will update the ExperimentRun row with the status (either failed or completed)
    """
    @wraps(entrypoint_func)
    def with_entrypoint(self, *args, **kwargs):
        entrypoint_name = entrypoint_func.__name__
        with get_run_for_update(self.db_engine, self.run_id) as run_obj:
            if not run_obj.start_method:
                run_obj.start_method = entrypoint_name
        try:
            return_value = entrypoint_func(self, *args, **kwargs)
        except Exception as exc:
            with get_run_for_update(self.db_engine, self.run_id) as run_obj:
                run_obj.current_status = ExperimentRunStatus.failed
                run_obj.stacktrace = str(exc)
            raise exc

        with get_run_for_update(self.db_engine, self.run_id) as run_obj:
            run_obj.current_status = ExperimentRunStatus.completed

        return return_value

    return with_entrypoint


def increment_field(field, run_id, db_engine):
    """Increment an ExperimentRun's named field.

    Expects that the field is an integer in the database.

    Will also kick the last_updated_time timestamp.

    Args:
        field (str) The name of the field
        run_id (int) The identifier/primary key of the run
        db_engine (sqlalchemy.engine)
    """
    with scoped_session(db_engine) as session:
        # Use an update query instead of a session merge so it happens in one atomic query
        # and protect against race conditions
        session.query(ExperimentRun).filter_by(run_id=run_id).update({
            field: getattr(ExperimentRun, field) + 1,
            'last_updated_time': datetime.datetime.now()
        })


def record_matrix_building_started(run_id, db_engine):
    """Mark the current timestamp as the time at which matrix building started

    Args:
        run_id (int) The identifier/primary key of the run
        db_engine (sqlalchemy.engine)
    """
    with get_run_for_update(db_engine, run_id) as run_obj:
        run_obj.matrix_building_started = datetime.datetime.now()


def record_model_building_started(run_id, db_engine):
    """Mark the current timestamp as the time at which model building started

    Args:
        run_id (int) The identifier/primary key of the run
        db_engine (sqlalchemy.engine)
    """
    with get_run_for_update(db_engine, run_id) as run_obj:
        run_obj.model_building_started = datetime.datetime.now()


def built_matrix(run_id, db_engine):
    """Increment the matrix build counter for the ExperimentRun

    Args:
        run_id (int) The identifier/primary key of the run
        db_engine (sqlalchemy.engine)
    """
    increment_field('matrices_made', run_id, db_engine)


def skipped_matrix(run_id, db_engine):
    """Increment the matrix skip counter for the ExperimentRun

    Args:
        run_id (int) The identifier/primary key of the run
        db_engine (sqlalchemy.engine)
    """
    increment_field('matrices_skipped', run_id, db_engine)


def errored_matrix(run_id, db_engine):
    """Increment the matrix error counter for the ExperimentRun

    Args:
        run_id (int) The identifier/primary key of the run
        db_engine (sqlalchemy.engine)
    """
    increment_field('matrices_errored', run_id, db_engine)


def built_model(run_id, db_engine):
    """Increment the model build counter for the ExperimentRun

    Args:
        run_id (int) The identifier/primary key of the run
        db_engine (sqlalchemy.engine)
    """
    increment_field('models_made', run_id, db_engine)


def skipped_model(run_id, db_engine):
    """Increment the model skip counter for the ExperimentRun

    Args:
        run_id (int) The identifier/primary key of the run
        db_engine (sqlalchemy.engine)
    """
    increment_field('models_skipped', run_id, db_engine)


def errored_model(run_id, db_engine):
    """Increment the model error counter for the ExperimentRun

    Args:
        run_id (int) The identifier/primary key of the run
        db_engine (sqlalchemy.engine)
    """
    increment_field('models_errored', run_id, db_engine)



"""empty message

Revision ID: 9bbfdcf8bab0
Revises: fa1760d35710
Create Date: 2020-07-19 01:04:23.442598

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '9bbfdcf8bab0'
down_revision = 'fa1760d35710'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('experiment_runs', sa.Column('random_seed', sa.Integer(), nullable=True), schema='triage_metadata')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('experiment_runs', 'random_seed', schema='triage_metadata')
    # ### end Alembic commands ###



from triage.experiments import ExperimentBase


class SingleThreadedExperiment(ExperimentBase):
    def process_query_tasks(self, query_tasks):
        self.feature_generator.process_table_tasks(query_tasks)

    def process_matrix_build_tasks(self, matrix_build_tasks):
        self.matrix_builder.build_all_matrices(matrix_build_tasks)

    def process_train_test_batches(self, batches):
        self.model_train_tester.process_all_batches(batches)

    def process_subset_tasks(self, subset_tasks):
        self.subsetter.process_all_tasks(subset_tasks)



"""add nuke triage function

Revision ID: a98acf92fd48
Revises: 4ae804cc0977
Create Date: 2020-07-19 01:46:02.751987

"""
from alembic import op
import os

# revision identifiers, used by Alembic.
revision = 'a98acf92fd48'
down_revision = '4ae804cc0977'
branch_labels = None
depends_on = None


def upgrade():
    nuke_triage_filename = os.path.join(
        os.path.dirname(__file__), "../../sql/nuke_triage.sql"
    )
    with open(nuke_triage_filename) as fd:
        stmt = fd.read()
        op.execute(stmt)



def downgrade():
    pass



import csv
import datetime
import hashlib
import numpy as np
import pandas as pd
import json

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import random
from itertools import chain
from functools import partial

import postgres_copy
import sqlalchemy
from retrying import retry
from sqlalchemy.orm import sessionmaker
from ohio import PipeTextIO

from triage.component.results_schema import (
    Experiment,
    Matrix,
    Model,
    ExperimentMatrix,
    ExperimentModel,
)


def filename_friendly_hash(inputs):
    def dt_handler(x):
        if isinstance(x, datetime.datetime) or isinstance(x, datetime.date):
            return x.isoformat()
        raise TypeError("Unknown type")

    return hashlib.md5(
        json.dumps(inputs, default=dt_handler, sort_keys=True).encode("utf-8")
    ).hexdigest()


def get_subset_table_name(subset_config):
    return "subset_{}_{}".format(
        subset_config.get("name", "default"),
        filename_friendly_hash(subset_config),
    )


def retry_if_db_error(exception):
    return isinstance(exception, sqlalchemy.exc.OperationalError)


DEFAULT_RETRY_KWARGS = {
    "retry_on_exception": retry_if_db_error,
    "wait_exponential_multiplier": 1000,  # wait 2^x*1000ms between each retry
    "stop_max_attempt_number": 14,
    # with this configuration, last wait will be ~2 hours
    # for a total of ~4.5 hours waiting
}


db_retry = retry(**DEFAULT_RETRY_KWARGS)


@db_retry
def save_experiment_and_get_hash(config, random_seed, db_engine):
    experiment_hash = filename_friendly_hash(config)
    session = sessionmaker(bind=db_engine)()
    session.merge(Experiment(experiment_hash=experiment_hash, random_seed=random_seed, config=config))
    session.commit()
    session.close()
    return experiment_hash


@db_retry
def associate_matrices_with_experiment(experiment_hash, matrix_uuids, db_engine):
    session = sessionmaker(bind=db_engine)()
    for matrix_uuid in matrix_uuids:
        session.merge(ExperimentMatrix(experiment_hash=experiment_hash, matrix_uuid=matrix_uuid))
    session.commit()
    session.close()
    logger.spam("Associated matrices with experiment in database")


@db_retry
def associate_models_with_experiment(experiment_hash, model_hashes, db_engine):
    session = sessionmaker(bind=db_engine)()
    for model_hash in model_hashes:
        session.merge(ExperimentModel(experiment_hash=experiment_hash, model_hash=model_hash))
    session.commit()
    session.close()
    logger.spam("Associated models with experiment in database")


@db_retry
def missing_matrix_uuids(experiment_hash, db_engine):
    """Compare the contents of the experiment_matrices table with that of the
    matrices table to produce a list of matrix_uuids that the experiment wants
    but are not available.
    """
    query = f"""
        select experiment_matrices.matrix_uuid
        from {ExperimentMatrix.__table__.fullname} experiment_matrices
        left join {Matrix.__table__.fullname} matrices
        on (experiment_matrices.matrix_uuid = matrices.matrix_uuid)
        where experiment_hash = %s
        and matrices.matrix_uuid is null
    """
    return [row[0] for row in db_engine.execute(query, experiment_hash)]


@db_retry
def missing_model_hashes(experiment_hash, db_engine):
    """Compare the contents of the experiment_models table with that of the
    models table to produce a list of model hashes the experiment wants
    but are not available.
    """
    query = f"""
        select experiment_models.model_hash
        from {ExperimentModel.__table__.fullname} experiment_models
        left join {Model.__table__.fullname} models
        on (experiment_models.model_hash = models.model_hash)
        where experiment_hash = %s
        and models.model_hash is null
    """
    return [row[0] for row in db_engine.execute(query, experiment_hash)]


class Batch:
    # modified from
    # http://codereview.stackexchange.com/questions/118883/split-up-an-iterable-into-batches
    def __init__(self, iterable, limit=None):
        self.iterator = iter(iterable)
        self.limit = limit
        try:
            self.current = next(self.iterator)
        except StopIteration:
            self.on_going = False
        else:
            self.on_going = True

    def group(self):
        yield self.current
        # start enumerate at 1 because we already yielded the last saved item
        for num, item in enumerate(self.iterator, 1):
            self.current = item
            if num == self.limit:
                break
            yield item
        else:
            self.on_going = False

    def __iter__(self):
        while self.on_going:
            yield self.group()


AVAILABLE_TIEBREAKERS = {'random', 'best', 'worst'}

def sort_predictions_and_labels(predictions_proba, labels, df_index, tiebreaker='random', sort_seed=None):
    """Sort predictions and labels with a configured tiebreaking rule

    Args:
        predictions_proba (np.array) The predicted scores
        labels (np.array) The numeric labels (1/0, not True/False)
        df_index (pd.MultiIndex) Index (generally entity_id, as_of_date tuples) to be sorted with the labels/scores
        tiebreaker (string) The tiebreaking method ('best', 'worst', 'random')
        sort_seed (signed int) The sort seed. Needed if 'random' tiebreaking is picked.

    Returns:
        (tuple) (predictions_proba, labels, df_index), sorted
    """
    if len(labels) == 0:
        logger.notice("No labels present, skipping predictions sorting .")
        return (predictions_proba, labels, df_index)

    df = pd.DataFrame(predictions_proba, columns=["score"])
    df['label_value'] = labels
    df.set_index(df_index, inplace=True)


    if tiebreaker == 'random':
        if not sort_seed:
            raise ValueError("If random tiebreaker is used, a sort seed must be given")
        random.seed(sort_seed)
        np.random.seed(sort_seed)
        df['random'] = np.random.rand(len(df))
        df.sort_values(by=['score', 'random'], inplace=True, ascending=[False, False])
        df.drop('random', axis=1)
    elif tiebreaker == 'worst':
        df.sort_values(by=["score", "label_value"], inplace=True, ascending=[False,True], na_position='first')
    elif tiebreaker == 'best':
         df.sort_values(by=["score", "label_value"], inplace=True, ascending=[False,False], na_position='last')
    else:
        raise ValueError(f"Unknown tiebreaker: {tiebreaker}")

    return  [
        df['score'].to_numpy(),
        df['label_value'].to_numpy(),
        df.index
    ]


@db_retry
def retrieve_model_id_from_hash(db_engine, model_hash):
    """Retrieves a model id from the database that matches the given hash

    Args:
        db_engine (sqlalchemy.engine) A database engine
        model_hash (str) The model hash to lookup

    Returns: (int) The model id (if found in DB), None (if not)
    """
    session = sessionmaker(bind=db_engine)()
    try:
        saved = session.query(Model).filter_by(model_hash=model_hash).one_or_none()
        return saved.model_id if saved else None
    finally:
        session.close()


@db_retry
def retrieve_model_hash_from_id(db_engine, model_id):
    """Retrieves the model hash associated with a given model id

    Args:
        model_id (int) The id of a given model in the database

    Returns: (str) the stored hash of the model
    """
    session = sessionmaker(bind=db_engine)()
    try:
        return session.query(Model).get(model_id).model_hash
    finally:
        session.close()


def _write_csv(file_like, db_objects, type_of_object):
    writer = csv.writer(file_like, quoting=csv.QUOTE_MINIMAL, lineterminator='\n')
    for db_object in db_objects:
        if type(db_object) != type_of_object:
            raise TypeError("Cannot copy collection of objects to db as they are not all "
                            f"of the same type. First object was {type_of_object} "
                            f"and later encountered a {type(db_object)}")
        writer.writerow(
            [getattr(db_object, col.name) for col in db_object.__table__.columns]
        )


@db_retry
def save_db_objects(db_engine, db_objects):
    """Saves a collection of SQLAlchemy model objects to the database using a COPY command

    Args:
        db_engine (sqlalchemy.engine)
        db_objects (iterable) SQLAlchemy model objects, corresponding to a valid table
    """
    db_objects = iter(db_objects)
    first_object = next(db_objects)
    type_of_object = type(first_object)
    columns = [col.name for col in first_object.__table__.columns]

    with PipeTextIO(partial(
            _write_csv,
            db_objects=chain((first_object,), db_objects),
            type_of_object=type_of_object
    )) as pipe:
        postgres_copy.copy_from(source=pipe, dest=type_of_object,
                                engine_or_conn=db_engine,
                                columns=columns,
                                format="csv")



# coding: utf-8

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from sklearn.linear_model import LogisticRegression

from .transformers import CutOff


class ScaledLogisticRegression(BaseEstimator, ClassifierMixin):
    """
    An in-place replacement for the scikit-learn's LogisticRegression.

    It incorporates the MaxMinScaler, and the CutOff as preparations
    for the  logistic regression.
    """

    def __init__(
        self,
        penalty="l2",
        dual=False,
        tol=1e-4,
        C=1.0,
        fit_intercept=True,
        intercept_scaling=1,
        class_weight=None,
        random_state=None,
        solver="saga",
        max_iter=100,
        multi_class="ovr",
        verbose=0,
        warm_start=False,
        n_jobs=1,
    ):
        self.penalty = penalty
        self.dual = dual
        self.tol = tol
        self.C = C
        self.fit_intercept = fit_intercept
        self.intercept_scaling = intercept_scaling
        self.class_weight = class_weight
        self.random_state = random_state
        self.solver = solver
        self.max_iter = max_iter
        self.multi_class = multi_class
        self.verbose = verbose
        self.warm_start = warm_start
        self.n_jobs = n_jobs

        self.minmax_scaler = MinMaxScaler()
        self.dsapp_cutoff = CutOff()
        self.lr = LogisticRegression(
            penalty=penalty,
            dual=dual,
            tol=tol,
            C=C,
            fit_intercept=fit_intercept,
            intercept_scaling=intercept_scaling,
            class_weight=class_weight,
            random_state=random_state,
            solver=solver,
            max_iter=max_iter,
            multi_class=multi_class,
            verbose=verbose,
            warm_start=warm_start,
            n_jobs=n_jobs,
        )

        self.pipeline = Pipeline(
            [
                ("minmax_scaler", self.minmax_scaler),
                ("dsapp_cutoff", self.dsapp_cutoff),
                ("lr", self.lr),
            ]
        )

    def fit(self, X, y=None):
        self.pipeline.fit(X, y)

        self.min_ = self.pipeline.named_steps["minmax_scaler"].min_
        self.scale_ = self.pipeline.named_steps["minmax_scaler"].scale_
        self.data_min_ = self.pipeline.named_steps["minmax_scaler"].data_min_
        self.data_max_ = self.pipeline.named_steps["minmax_scaler"].data_max_
        self.data_range_ = self.pipeline.named_steps["minmax_scaler"].data_range_

        self.coef_ = self.pipeline.named_steps["lr"].coef_
        self.intercept_ = self.pipeline.named_steps["lr"].intercept_

        self.classes_ = self.pipeline.named_steps["lr"].classes_

        return self

    def predict_proba(self, X):
        return self.pipeline.predict_proba(X)

    def predict_log_proba(self, X):
        return self.pipeline.predict_log_proba(X)

    def predict(self, X):
        return self.pipeline.predict(X)

    def score(self, X, y):
        return self.pipeline.score(X, y)



from triage.component.postmodeling.contrast.model_evaluator import ModelEvaluator
from triage.component.postmodeling.crosstabs import run_crosstabs
from tests.utils import sample_config, populate_source_data, assert_plot_figures_added
from triage.experiments import SingleThreadedExperiment
import pandas as pd
import pytest


@pytest.fixture(scope="module")
def model_evaluator(shared_db_engine, shared_project_storage):
    """Returns an instantiated ModelEvaluator available at module scope"""
    populate_source_data(shared_db_engine)
    base_config = sample_config()
    # We need to have an ensemble model to test ModelEvaluator correctly
    # so we can't use the finished_experiment fixture"""
    base_config["grid_config"] = {
        "sklearn.ensemble.ExtraTreesClassifier": {
            "n_estimators": [10],
            "criterion": ["gini"],
            "max_depth": [1],
            "max_features": ["sqrt"],
            "min_samples_split": [2],
        }
    }
    SingleThreadedExperiment(
        base_config,
        db_engine=shared_db_engine,
        project_path=shared_project_storage.project_path,
    ).run()
    return ModelEvaluator(1, 1, shared_db_engine)


def test_ModelEvaluator_model_type(model_evaluator):
    assert model_evaluator.model_type == "sklearn.ensemble.ExtraTreesClassifier"


def test_ModelEvaluator_predictions(model_evaluator):
    assert isinstance(model_evaluator.predictions, pd.DataFrame)


def test_ModelEvaluator_feature_importances(model_evaluator):
    assert isinstance(model_evaluator.feature_importances(), pd.DataFrame)


def test_ModelEvaluator_feature_group_importances(model_evaluator):
    assert isinstance(model_evaluator.feature_group_importances(), pd.DataFrame)


def test_ModelEvaluator_test_metrics(model_evaluator):
    assert isinstance(model_evaluator.test_metrics, pd.DataFrame)


def test_ModelEvaluator_train_metrics(model_evaluator):
    assert isinstance(model_evaluator.train_metrics, pd.DataFrame)


def test_ModelEvaluator_crosstabs(model_evaluator, crosstabs_config):
    run_crosstabs(model_evaluator.engine, crosstabs_config)
    assert isinstance(model_evaluator.crosstabs, pd.DataFrame)


def test_ModelEvaluator_preds_matrix(model_evaluator, shared_project_storage):
    assert isinstance(
        model_evaluator.preds_matrix(shared_project_storage.project_path), pd.DataFrame
    )


def test_ModelEvaluator_plot_score_distribution(model_evaluator):
    with assert_plot_figures_added():
        model_evaluator.plot_score_distribution()


def test_ModelEvaluator_plot_score_label_distributions(model_evaluator):
    with assert_plot_figures_added():
        model_evaluator.plot_score_label_distributions()


def test_ModelEvaluator_plot_score_distribution_thresh(model_evaluator):
    with assert_plot_figures_added():
        model_evaluator.plot_score_distribution_thresh(param_type="rank_abs", param=10)


def test_ModelEvaluator_plot_feature_importances(
    model_evaluator, shared_project_storage
):
    with assert_plot_figures_added():
        model_evaluator.plot_feature_importances(shared_project_storage.project_path)


def test_ModelEvaluator_plot_feature_importances_std_err(
    model_evaluator, shared_project_storage
):
    with assert_plot_figures_added():
        model_evaluator.plot_feature_importances_std_err(
            shared_project_storage.project_path
        )


def test_ModelEvaluator_plot_precision_recall_n(model_evaluator):
    with assert_plot_figures_added():
        model_evaluator.plot_precision_recall_n()


def test_ModelEvaluator_plot_recall_fpr_n(model_evaluator):
    with assert_plot_figures_added():
        model_evaluator.plot_recall_fpr_n()


def test_ModelEvaluator_plot_ROC(model_evaluator):
    with assert_plot_figures_added():
        model_evaluator.plot_ROC()


def test_ModelEvaluator_compute_AUC(model_evaluator):
    assert isinstance(model_evaluator.compute_AUC(), tuple)


def test_ModelEvaluator_cluster_correlation_sparsity(
    model_evaluator, shared_project_storage
):
    with assert_plot_figures_added():
        model_evaluator.cluster_correlation_sparsity(
            shared_project_storage.project_path
        )


def test_ModelEvaluator_cluster_correlation_features(
    model_evaluator, shared_project_storage
):
    with assert_plot_figures_added():
        model_evaluator.cluster_correlation_features(
            shared_project_storage.project_path
        )


def test_ModelEvaluator_plot_feature_group_aggregate_importances(
    model_evaluator, shared_project_storage
):
    with assert_plot_figures_added():
        model_evaluator.plot_feature_group_aggregate_importances(
            path=shared_project_storage.project_path
        )



from triage.component.collate import Categorical, Compare


def assert_contains(haystack, needle):
    for h in haystack:
        if type(h) is tuple and needle in h[0]:
            return
        elif needle in h:
            return
    assert False


def test_compare_lists():
    d = Compare("col", "=", ["a", "b", "c"], [], {}, include_null=True).quantities
    assert len(d) == 4
    assert len(set(d.values())) == len(d)
    assert len(set(d.keys())) == len(d)
    assert_contains(d.values(), "col = 'a'")
    assert_contains(d.values(), "col = 'b'")
    assert_contains(d.values(), "col = 'c'")
    assert_contains(map(lambda x: x[0].lower(), d.values()), "col is null")

    d = Compare("col", ">", [1, 2, 3], [], {}).quantities
    assert len(d) == 3
    assert len(set(d.values())) == len(d)
    assert len(set(d.keys())) == len(d)
    assert_contains(d.values(), "col > 1")
    assert_contains(d.values(), "col > 2")
    assert_contains(d.values(), "col > 3")

    d = Compare("col", "=", ["a", "b", "c"], [], {}, include_null=False).quantities
    assert len(d) == 3
    assert len(set(d.values())) == len(d)
    assert len(set(d.keys())) == len(d)
    assert_contains(d.values(), "col = 'a'")
    assert_contains(d.values(), "col = 'b'")
    assert_contains(d.values(), "col = 'c'")

    d = Compare(
        "really_long_column_name",
        "=",
        [
            "really long string value that is similar to others",
            "really long string value that is like others",
            "really long string value that is quite alike to others",
            "really long string value that is also like everything else",
        ],
        [],
        {},
        maxlen=32,
    ).quantities
    assert len(d) == 4
    assert len(set(d.values())) == len(d)
    assert len(set(d.keys())) == len(d)
    assert all(len(k) <= 32 for k in d.keys())
    assert_contains(
        d.values(),
        "really_long_column_name = 'really long string value that is similar to others'",
    )
    assert_contains(
        d.values(),
        "really_long_column_name = 'really long string value that is like others'",
    )
    assert_contains(
        d.values(),
        "really_long_column_name = 'really long string value that is quite alike to others'",
    )
    assert_contains(
        d.values(),
        "really_long_column_name = 'really long string value that is also like everything else'",
    )


def test_compare_override_quoting():
    d = Compare(
        "col",
        "@>",
        {"one": "array['one'::varchar]", "two": "array['two'::varchar]"},
        [],
        {},
        quote_choices=False,
    ).quantities
    assert len(d) == 2
    assert_contains(d.values(), "col @> array['one'::varchar]")
    assert_contains(d.values(), "col @> array['two'::varchar]")


def test_compare_dicts():
    d = Compare(
        "col", "=", {"vala": "a", "valb": "b", "valc": "c"}, [], {}, include_null=True
    ).quantities
    assert len(d) == 4
    assert len(set(d.values())) == len(d)
    assert len(set(d.keys())) == len(d)
    assert_contains(d.values(), "col = 'a'")
    assert_contains(d.values(), "col = 'b'")
    assert_contains(d.values(), "col = 'c'")
    assert_contains(d.keys(), "vala")
    assert_contains(d.keys(), "valb")
    assert_contains(d.keys(), "valc")
    assert_contains(map(str.lower, d.keys()), "null")
    assert_contains(map(lambda x: x[0].lower(), d.values()), "col is null")

    d = Compare(
        "col", "<", {"val1": 1, "val2": 2, "val3": 3}, [], {}, include_null="missing"
    ).quantities
    assert len(d) == 4
    assert len(set(d.values())) == len(d)
    assert len(set(d.keys())) == len(d)
    assert_contains(d.values(), "col < 1")
    assert_contains(d.values(), "col < 2")
    assert_contains(d.values(), "col < 3")
    assert_contains(map(lambda x: x[0].lower(), d.values()), "null")
    assert_contains(d.keys(), "val1")
    assert_contains(d.keys(), "val2")
    assert_contains(d.keys(), "val3")
    assert_contains(d.keys(), "missing")

    d = Compare(
        "long_column_name",
        "=",
        {
            "really long string key that is similar to others":
                "really long string value that is similar to others",
            "really long string key that is like others":
                "really long string value that is like others",
            "different key": "really long string value that is quite alike to others",
            "ni": "really long string value that is also like everything else",
        },
        [],
        {},
        maxlen=32,
    ).quantities
    assert len(d) == 4
    assert len(set(d.values())) == len(d)
    assert len(set(d.keys())) == len(d)
    assert all(len(k) <= 32 for k in d.keys())
    assert_contains(d.keys(), "differ")
    assert_contains(
        d.values(),
        "long_column_name = 'really long string value that is similar to others'",
    )
    assert_contains(
        d.values(), "long_column_name = 'really long string value that is like others'"
    )
    assert_contains(
        d.values(),
        "long_column_name = 'really long string value that is quite alike to others'",
    )
    assert_contains(
        d.values(),
        "long_column_name = 'really long string value that is also like everything else'",
    )


def test_categorical_same_as_compare():
    d1 = Categorical("col", {"vala": "a", "valb": "b", "valc": "c"}, [], {}).quantities
    d2 = Compare("col", "=", {"vala": "a", "valb": "b", "valc": "c"}, [], {}).quantities
    assert sorted(d1.values()) == sorted(d2.values())
    d3 = Categorical(
        "col", {"vala": "a", "valb": "b", "valc": "c"}, [], {}, op_in_name=True
    ).quantities
    assert d2 == d3


def test_categorical_nones():
    d1 = Categorical(
        "col", {"vala": "a", "valb": "b", "valc": "c", "_NULL": None}, [], {}
    ).quantities
    d2 = Compare(
        "col",
        "=",
        {"vala": "a", "valb": "b", "valc": "c"},
        [],
        {},
        op_in_name=False,
        include_null=True,
    ).quantities
    assert d1 == d2
    d3 = Categorical("col", ["a", "b", "c", None], [], {}).quantities
    assert sorted(d1.values()) == sorted(d3.values())



from datetime import datetime, timedelta
from unittest.mock import patch

import factory
import numpy as np
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition.distance_from_best import (
    DistanceFromBestTable,
    BestDistancePlotter,
)
from triage.component.catwalk.db import ensure_db

from tests.results_tests.factories import (
    EvaluationFactory,
    ModelFactory,
    ModelGroupFactory,
    init_engine,
    session,
)

from .utils import create_sample_distance_table


def _sql_add_days(sql_date, days):
    return datetime.strftime(
        datetime.strptime(sql_date, "%Y-%m-%d") + timedelta(days=days), "%Y-%m-%d"
    )


def test_DistanceFromBestTable():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ensure_db(engine)
        init_engine(engine)
        model_groups = {
            "stable": ModelGroupFactory(model_type="myStableClassifier"),
            "bad": ModelGroupFactory(model_type="myBadClassifier"),
            "spiky": ModelGroupFactory(model_type="mySpikeClassifier"),
        }

        class StableModelFactory(ModelFactory):
            model_group_rel = model_groups["stable"]

        class BadModelFactory(ModelFactory):
            model_group_rel = model_groups["bad"]

        class SpikyModelFactory(ModelFactory):
            model_group_rel = model_groups["spiky"]

        models = {
            "stable_3y_ago": StableModelFactory(train_end_time="2014-01-01"),
            "stable_2y_ago": StableModelFactory(train_end_time="2015-01-01"),
            "stable_1y_ago": StableModelFactory(train_end_time="2016-01-01"),
            "bad_3y_ago": BadModelFactory(train_end_time="2014-01-01"),
            "bad_2y_ago": BadModelFactory(train_end_time="2015-01-01"),
            "bad_1y_ago": BadModelFactory(train_end_time="2016-01-01"),
            "spiky_3y_ago": SpikyModelFactory(train_end_time="2014-01-01"),
            "spiky_2y_ago": SpikyModelFactory(train_end_time="2015-01-01"),
            "spiky_1y_ago": SpikyModelFactory(train_end_time="2016-01-01"),
        }

        class ImmediateEvalFactory(EvaluationFactory):
            evaluation_start_time = factory.LazyAttribute(
                lambda o: o.model_rel.train_end_time
            )
            evaluation_end_time = factory.LazyAttribute(
                lambda o: _sql_add_days(o.model_rel.train_end_time, 1)
            )

        class MonthOutEvalFactory(EvaluationFactory):
            evaluation_start_time = factory.LazyAttribute(
                lambda o: _sql_add_days(o.model_rel.train_end_time, 31)
            )
            evaluation_end_time = factory.LazyAttribute(
                lambda o: _sql_add_days(o.model_rel.train_end_time, 32)
            )

        class Precision100Factory(ImmediateEvalFactory):
            metric = "precision@"
            parameter = "100_abs"

        class Precision100FactoryMonthOut(MonthOutEvalFactory):
            metric = "precision@"
            parameter = "100_abs"

        class Recall100Factory(ImmediateEvalFactory):
            metric = "recall@"
            parameter = "100_abs"

        class Recall100FactoryMonthOut(MonthOutEvalFactory):
            metric = "recall@"
            parameter = "100_abs"

        for (add_val, PrecFac, RecFac) in (
            (0, Precision100Factory, Recall100Factory),
            (-0.15, Precision100FactoryMonthOut, Recall100FactoryMonthOut),
        ):
            PrecFac(model_rel=models["stable_3y_ago"], stochastic_value=0.6 + add_val)
            PrecFac(model_rel=models["stable_2y_ago"], stochastic_value=0.57 + add_val)
            PrecFac(model_rel=models["stable_1y_ago"], stochastic_value=0.59 + add_val)
            PrecFac(model_rel=models["bad_3y_ago"], stochastic_value=0.4 + add_val)
            PrecFac(model_rel=models["bad_2y_ago"], stochastic_value=0.39 + add_val)
            PrecFac(model_rel=models["bad_1y_ago"], stochastic_value=0.43 + add_val)
            PrecFac(model_rel=models["spiky_3y_ago"], stochastic_value=0.8 + add_val)
            PrecFac(model_rel=models["spiky_2y_ago"], stochastic_value=0.4 + add_val)
            PrecFac(model_rel=models["spiky_1y_ago"], stochastic_value=0.4 + add_val)
            RecFac(model_rel=models["stable_3y_ago"], stochastic_value=0.55 + add_val)
            RecFac(model_rel=models["stable_2y_ago"], stochastic_value=0.56 + add_val)
            RecFac(model_rel=models["stable_1y_ago"], stochastic_value=0.55 + add_val)
            RecFac(model_rel=models["bad_3y_ago"], stochastic_value=0.35 + add_val)
            RecFac(model_rel=models["bad_2y_ago"], stochastic_value=0.34 + add_val)
            RecFac(model_rel=models["bad_1y_ago"], stochastic_value=0.36 + add_val)
            RecFac(model_rel=models["spiky_3y_ago"], stochastic_value=0.35 + add_val)
            RecFac(model_rel=models["spiky_2y_ago"], stochastic_value=0.8 + add_val)
            RecFac(model_rel=models["spiky_1y_ago"], stochastic_value=0.36 + add_val)
        session.commit()
        distance_table = DistanceFromBestTable(
            db_engine=engine, models_table="models", distance_table="dist_table", agg_type="worst"
        )
        metrics = [
            {"metric": "precision@", "parameter": "100_abs"},
            {"metric": "recall@", "parameter": "100_abs"},
        ]
        model_group_ids = [mg.model_group_id for mg in model_groups.values()]
        distance_table.create_and_populate(
            model_group_ids, ["2014-01-01", "2015-01-01", "2016-01-01"], metrics
        )

        # get an ordered list of the model groups for a particular metric/time
        query = """
            select model_group_id, raw_value, dist_from_best_case, dist_from_best_case_next_time
            from dist_table where metric = %s and parameter = %s and train_end_time = %s
            order by dist_from_best_case
        """

        prec_3y_ago = engine.execute(query, ("precision@", "100_abs", "2014-01-01"))
        assert [row for row in prec_3y_ago] == [
            (models["spiky_3y_ago"].model_group_id, 0.8, 0, 0.17),
            (models["stable_3y_ago"].model_group_id, 0.6, 0.2, 0),
            (models["bad_3y_ago"].model_group_id, 0.4, 0.4, 0.18),
        ]

        recall_2y_ago = engine.execute(query, ("recall@", "100_abs", "2015-01-01"))
        assert [row for row in recall_2y_ago] == [
            (models["spiky_2y_ago"].model_group_id, 0.8, 0, 0.19),
            (models["stable_2y_ago"].model_group_id, 0.56, 0.24, 0),
            (models["bad_2y_ago"].model_group_id, 0.34, 0.46, 0.19),
        ]

        assert distance_table.observed_bounds == {
            ("precision@", "100_abs"): (0.39, 0.8),
            ("recall@", "100_abs"): (0.34, 0.8),
        }


def test_BestDistancePlotter():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)
        plotter = BestDistancePlotter(distance_table)
        df_dist = plotter.generate_plot_data(
            metric="precision@",
            parameter="100_abs",
            model_group_ids=[1, 2],
            train_end_times=["2014-01-01", "2015-01-01"],
        )
        # assert that we have the right # of columns and a row for each % diff value
        # 202 row because 101 percentiles (0-100 inclusive), 2 model groups
        assert df_dist.shape == (101 * 2, 5)

        # all of the model groups are within .34 of the best, so pick
        # a number higher than that and all should qualify
        for value in df_dist[df_dist["distance"] == 0.35]["pct_of_time"].values:
            assert np.isclose(value, 1.0)

        # model group 1 (stable) should be within 0.11 1/2 of the time
        # if we included 2016 in the train_end_times, this would be 1/3!
        for value in df_dist[
            (df_dist["distance"] == 0.11) & (df_dist["model_group_id"] == 1)
        ]["pct_of_time"].values:
            assert np.isclose(value, 0.5)


def test_BestDistancePlotter_plot():
    with patch("triage.component.audition.distance_from_best.plot_cats") as plot_patch:
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            distance_table, model_groups = create_sample_distance_table(engine)
            plotter = BestDistancePlotter(distance_table)
            plotter.plot_all_best_dist(
                [{"metric": "precision@", "parameter": "100_abs"}],
                model_group_ids=[1, 2],
                train_end_times=["2014-01-01", "2015-01-01"],
            )
        assert plot_patch.called
        args, kwargs = plot_patch.call_args
        assert "distance" in kwargs["frame"]
        assert "pct_of_time" in kwargs["frame"]
        assert kwargs["x_col"] == "distance"
        assert kwargs["y_col"] == "pct_of_time"


def test_BestDistancePlotter_plot_bounds():
    class FakeDistanceTable:
        @property
        def observed_bounds(self):
            return {
                ("precision@", "100_abs"): (0.02, 0.87),
                ("recall@", "100_abs"): (0.0, 1.0),
                ("false positives@", "300_abs"): (2, 162),
            }

    plotter = BestDistancePlotter(FakeDistanceTable())
    assert plotter.plot_bounds("precision@", "100_abs") == (0.0, 1.0)
    assert plotter.plot_bounds("recall@", "100_abs") == (0.0, 1.0)
    assert plotter.plot_bounds("false positives@", "300_abs") == (2, 178)



from triage.component.catwalk.model_trainers import NO_FEATURE_IMPORTANCE


def _entity_feature_values(matrix, feature_name, as_of_date):
    """Finds the value of the given feature for each entity in a matrix

    Args:
        matrix (pandas.DataFrame), with index either 'entity_id' or 'entity_id'/'as_of_date'
        feature_name (string) The name of a column in the matrix
        as_of_date (datetime.date) This must be one of the valid as_of_dates in the matrix

    Returns: (list) of (entity_id, feature_value) tuples
    """
    results = []
    index_of_entity = matrix.index.names.index("entity_id")
    index_of_date = matrix.index.names.index("as_of_date")
    if feature_name == NO_FEATURE_IMPORTANCE:
        zipped_iter = zip(matrix.index.values, [None] * len(matrix.index.values))
    else:
        zipped_iter = zip(matrix.index.values, matrix[feature_name].tolist())
    for row in zipped_iter:
        index_values, feature_value = row
        entity_id = index_values[index_of_entity]
        if type(index_values[index_of_date].date()) != type(as_of_date):
            raise TypeError("Types of date in matrix and input must match, "
                            f"Matrix was {type(index_values[index_of_date].date())}",
                            f"Input was {type(as_of_date)}")
        if index_values[index_of_date].date() == as_of_date:
            results.append((entity_id, feature_value))
    return results


def uniform_distribution(db_engine, model_id, as_of_date, test_matrix_store, n_ranks):
    """Calculates individual feature importances based on the global feature importances

    Args:
        db_engine (sqlalchemy.engine)
        model_id (int) A model id, expected to be present in triage_metadata.models
        as_of_date (datetime.date) The date to produce individual importances as of
        test_matrix_store (catwalk.storage.MatrixStore) The test matrix
        n_ranks (int) Number of ranks to calculate and save. Defaults to 5

    Returns: (list) dicts with entity_id, feature_value, feature_name, score
    """
    global_feature_importances = [
        row
        for row in db_engine.execute(
            """select feature, feature_importance
        from train_results.feature_importances where model_id = %s
        order by feature_importance desc limit %s""",
            model_id,
            n_ranks,
        )
    ]

    results = []

    for feature_name, feature_importance in global_feature_importances:
        efv = _entity_feature_values(test_matrix_store.design_matrix, feature_name, as_of_date)
        for entity_id, feature_value in efv:
            results.append(
                {
                    "entity_id": entity_id,
                    "feature_value": feature_value,
                    "feature_name": feature_name,
                    "score": feature_importance,
                }
            )
    return results



import random


FLOAT_TO_INT_MULTIPLIER = 2000000000


def generate_python_random_seed():
    """Generate a random integer suitable for seeding the Python random generator
    """
    return int(random.uniform(0, 1.0) * FLOAT_TO_INT_MULTIPLIER)



import datetime
from unittest import TestCase

from triage.util.conf import convert_str_to_relativedelta

from triage.component.timechop import Timechop


class test_calculate_train_test_split_times(TestCase):
    def test_valid_input(self):
        expected_result = [
            datetime.datetime(2015, 3, 1, 0, 0),
            datetime.datetime(2015, 6, 1, 0, 0),
            datetime.datetime(2015, 9, 1, 0, 0),
            datetime.datetime(2015, 12, 1, 0, 0),
            datetime.datetime(2016, 3, 1, 0, 0),
            datetime.datetime(2016, 6, 1, 0, 0),
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2017, 1, 1, 0, 0),
            label_start_time=datetime.datetime(2015, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2017, 1, 1, 0, 0),
            model_update_frequency="3 months",
            training_as_of_date_frequencies=["1 day"],
            test_as_of_date_frequencies=["1 day"],
            max_training_histories=["1 year"],
            test_durations=["6 months"],
            test_label_timespans=["1 months"],
            training_label_timespans=["3 days"],
        )

        # this should throw an exception because last possible label date is after
        # end of feature time
        result = chopper.calculate_train_test_split_times(
            training_label_timespan=convert_str_to_relativedelta("3 days"),
            test_duration="6 months",
            test_label_timespan=convert_str_to_relativedelta("1 month"),
        )

        assert result == expected_result

    def test_labels_after_features(self):
        chopper = Timechop(
            feature_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2016, 1, 1, 0, 0),
            label_start_time=datetime.datetime(2015, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2017, 1, 1, 0, 0),
            model_update_frequency="3 months",
            training_as_of_date_frequencies=["1 day"],
            test_as_of_date_frequencies=["1 day"],
            max_training_histories=["1 year"],
            test_durations=["6 months"],
            test_label_timespans=["1 months"],
            training_label_timespans=["3 days"],
        )

        # this should throw an exception because last possible label date is after
        # end of feature time
        with self.assertRaises(ValueError):
            chopper.calculate_train_test_split_times(
                training_label_timespan=convert_str_to_relativedelta("3 days"),
                test_duration="6 months",
                test_label_timespan=convert_str_to_relativedelta("1 month"),
            )

    def test_no_valid_label_dates(self):
        chopper = Timechop(
            feature_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2016, 1, 1, 0, 0),
            label_start_time=datetime.datetime(2015, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2015, 2, 1, 0, 0),
            model_update_frequency="3 months",
            training_as_of_date_frequencies=["1 day"],
            test_as_of_date_frequencies=["1 day"],
            max_training_histories=["1 year"],
            test_durations=["6 months"],
            test_label_timespans=["1 months"],
            training_label_timespans=["3 days"],
        )

        # this should raise an error because there are no valid label dates in
        # the labeling time (label span is longer than labeling time)
        with self.assertRaises(ValueError):
            chopper.calculate_train_test_split_times(
                training_label_timespan=convert_str_to_relativedelta("3 days"),
                test_duration="6 months",
                test_label_timespan=convert_str_to_relativedelta("1 month"),
            )


def test_calculate_as_of_times_one_day_freq():
    expected_result = [
        datetime.datetime(2011, 1, 1, 0, 0),
        datetime.datetime(2011, 1, 2, 0, 0),
        datetime.datetime(2011, 1, 3, 0, 0),
        datetime.datetime(2011, 1, 4, 0, 0),
        datetime.datetime(2011, 1, 5, 0, 0),
        datetime.datetime(2011, 1, 6, 0, 0),
        datetime.datetime(2011, 1, 7, 0, 0),
        datetime.datetime(2011, 1, 8, 0, 0),
        datetime.datetime(2011, 1, 9, 0, 0),
        datetime.datetime(2011, 1, 10, 0, 0),
        datetime.datetime(2011, 1, 11, 0, 0),
    ]
    chopper = Timechop(
        feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
        feature_end_time=datetime.datetime(2012, 1, 1, 0, 0),
        label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
        label_end_time=datetime.datetime(2012, 1, 1, 0, 0),
        model_update_frequency="1 year",
        training_as_of_date_frequencies=["1 days"],
        test_as_of_date_frequencies=["7 days"],
        max_training_histories=["10 days", "1 year"],
        test_durations=["1 month"],
        test_label_timespans=["1 day"],
        training_label_timespans=["3 months"],
    )
    result = chopper.calculate_as_of_times(
        as_of_start_limit=datetime.datetime(2011, 1, 1, 0, 0),
        as_of_end_limit=datetime.datetime(2011, 1, 11, 0, 0),
        data_frequency=convert_str_to_relativedelta("1 days"),
    )
    assert result == expected_result


def test_calculate_as_of_times_three_day_freq():
    expected_result = [
        datetime.datetime(2011, 1, 1, 0, 0),
        datetime.datetime(2011, 1, 4, 0, 0),
        datetime.datetime(2011, 1, 7, 0, 0),
        datetime.datetime(2011, 1, 10, 0, 0),
    ]
    chopper = Timechop(
        feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
        feature_end_time=datetime.datetime(2012, 1, 1, 0, 0),
        label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
        label_end_time=datetime.datetime(2012, 1, 1, 0, 0),
        model_update_frequency="1 year",
        training_as_of_date_frequencies=["1 days"],
        test_as_of_date_frequencies=["7 days"],
        max_training_histories=["10 days", "1 year"],
        test_durations=["1 month"],
        test_label_timespans=["1 day"],
        training_label_timespans=["3 months"],
    )
    result = chopper.calculate_as_of_times(
        as_of_start_limit=datetime.datetime(2011, 1, 1, 0, 0),
        as_of_end_limit=datetime.datetime(2011, 1, 11, 0, 0),
        data_frequency=convert_str_to_relativedelta("3 days"),
        forward=True,
    )
    assert result == expected_result


class test_generate_matrix_definitions(TestCase):
    def test_look_back_time_equal_modeling_start(self):
        # TODO: rework this test since the test label window of 3 months
        # cannot be satisfied by the 10 day difference between modeling
        # start and end times, so it's not a very realistic case
        expected_result = {
            "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
            "label_start_time": datetime.datetime(2010, 1, 1, 0, 0),
            "feature_end_time": datetime.datetime(2010, 1, 11, 0, 0),
            "label_end_time": datetime.datetime(2010, 1, 11, 0, 0),
            "train_matrix": {
                "first_as_of_time": datetime.datetime(2010, 1, 1, 0, 0),
                "last_as_of_time": datetime.datetime(2010, 1, 5, 0, 0),
                "matrix_info_end_time": datetime.datetime(2010, 1, 6, 0, 0),
                "as_of_times": [
                    datetime.datetime(2010, 1, 1, 0, 0),
                    datetime.datetime(2010, 1, 2, 0, 0),
                    datetime.datetime(2010, 1, 3, 0, 0),
                    datetime.datetime(2010, 1, 4, 0, 0),
                    datetime.datetime(2010, 1, 5, 0, 0),
                ],
                "training_label_timespan": "1 day",
                "training_as_of_date_frequency": "1 days",
                "max_training_history": "5 days",
            },
            "test_matrices": [
                {
                    "first_as_of_time": datetime.datetime(2010, 1, 6, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 9, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 10, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0),
                    ],
                    "test_label_timespan": "1 day",
                    "test_as_of_date_frequency": "3 days",
                    "test_duration": "5 days",
                }
            ],
        }
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 11, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 11, 0, 0),
            model_update_frequency="5 days",
            training_as_of_date_frequencies=["1 days"],
            test_as_of_date_frequencies=["3 days"],
            max_training_histories=["5 days"],
            test_durations=["5 days"],
            test_label_timespans=["1 day"],
            training_label_timespans=["1 day"],
        )
        result = chopper.generate_matrix_definitions(
            train_test_split_time=datetime.datetime(2010, 1, 6, 0, 0),
            training_as_of_date_frequency="1 days",
            max_training_history="5 days",
            test_duration="5 days",
            test_label_timespan="1 day",
            training_label_timespan="1 day",
        )
        assert result == expected_result

    def test_look_back_time_before_modeling_start(self):
        expected_result = {
            "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
            "label_start_time": datetime.datetime(2010, 1, 1, 0, 0),
            "feature_end_time": datetime.datetime(2010, 1, 11, 0, 0),
            "label_end_time": datetime.datetime(2010, 1, 11, 0, 0),
            "train_matrix": {
                "first_as_of_time": datetime.datetime(2010, 1, 1, 0, 0),
                "last_as_of_time": datetime.datetime(2010, 1, 5, 0, 0),
                "matrix_info_end_time": datetime.datetime(2010, 1, 6, 0, 0),
                "as_of_times": [
                    datetime.datetime(2010, 1, 1, 0, 0),
                    datetime.datetime(2010, 1, 2, 0, 0),
                    datetime.datetime(2010, 1, 3, 0, 0),
                    datetime.datetime(2010, 1, 4, 0, 0),
                    datetime.datetime(2010, 1, 5, 0, 0),
                ],
                "training_label_timespan": "1 day",
                "training_as_of_date_frequency": "1 days",
                "max_training_history": "10 days",
            },
            "test_matrices": [
                {
                    "first_as_of_time": datetime.datetime(2010, 1, 6, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 9, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 10, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0),
                    ],
                    "test_label_timespan": "1 day",
                    "test_as_of_date_frequency": "3 days",
                    "test_duration": "5 days",
                },
                {
                    "first_as_of_time": datetime.datetime(2010, 1, 6, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 6, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 7, 0, 0),
                    "as_of_times": [datetime.datetime(2010, 1, 6, 0, 0)],
                    "test_label_timespan": "1 day",
                    "test_as_of_date_frequency": "6 days",
                    "test_duration": "5 days",
                },
            ],
        }
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 11, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 11, 0, 0),
            model_update_frequency="5 days",
            training_as_of_date_frequencies=["1 days"],
            test_as_of_date_frequencies=["3 days", "6 days"],
            max_training_histories=["10 days"],
            test_durations=["5 days"],
            test_label_timespans=["1 day"],
            training_label_timespans=["1 day"],
        )
        result = chopper.generate_matrix_definitions(
            train_test_split_time=datetime.datetime(2010, 1, 6, 0, 0),
            training_as_of_date_frequency="1 days",
            max_training_history="10 days",
            test_duration="5 days",
            test_label_timespan="1 day",
            training_label_timespan="1 day",
        )
        assert result == expected_result


class test_chop_time(TestCase):
    def test_evenly_divisible_values(self):
        expected_result = [
            {
                "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
                "label_start_time": datetime.datetime(2010, 1, 1, 0, 0),
                "feature_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "label_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "train_matrix": {
                    "first_as_of_time": datetime.datetime(2010, 1, 1, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 4, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 5, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 1, 0, 0),
                        datetime.datetime(2010, 1, 2, 0, 0),
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0),
                    ],
                    "training_label_timespan": "1 day",
                    "training_as_of_date_frequency": "1 days",
                    "max_training_history": "5 days",
                },
                "test_matrices": [
                    {
                        "first_as_of_time": datetime.datetime(2010, 1, 5, 0, 0),
                        "last_as_of_time": datetime.datetime(2010, 1, 9, 0, 0),
                        "matrix_info_end_time": datetime.datetime(2010, 1, 10, 0, 0),
                        "as_of_times": [
                            datetime.datetime(2010, 1, 5, 0, 0),
                            datetime.datetime(2010, 1, 6, 0, 0),
                            datetime.datetime(2010, 1, 7, 0, 0),
                            datetime.datetime(2010, 1, 8, 0, 0),
                            datetime.datetime(2010, 1, 9, 0, 0),
                        ],
                        "test_label_timespan": "1 day",
                        "test_as_of_date_frequency": "1 days",
                        "test_duration": "5 days",
                    }
                ],
            },
            {
                "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
                "label_start_time": datetime.datetime(2010, 1, 1, 0, 0),
                "feature_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "label_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "train_matrix": {
                    "first_as_of_time": datetime.datetime(2010, 1, 4, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 9, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 10, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 4, 0, 0),
                        datetime.datetime(2010, 1, 5, 0, 0),
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0),
                    ],
                    "training_label_timespan": "1 day",
                    "training_as_of_date_frequency": "1 days",
                    "max_training_history": "5 days",
                },
                "test_matrices": [
                    {
                        "first_as_of_time": datetime.datetime(2010, 1, 10, 0, 0),
                        "last_as_of_time": datetime.datetime(2010, 1, 14, 0, 0),
                        "matrix_info_end_time": datetime.datetime(2010, 1, 15, 0, 0),
                        "as_of_times": [
                            datetime.datetime(2010, 1, 10, 0, 0),
                            datetime.datetime(2010, 1, 11, 0, 0),
                            datetime.datetime(2010, 1, 12, 0, 0),
                            datetime.datetime(2010, 1, 13, 0, 0),
                            datetime.datetime(2010, 1, 14, 0, 0),
                        ],
                        "test_label_timespan": "1 day",
                        "test_as_of_date_frequency": "1 days",
                        "test_duration": "5 days",
                    }
                ],
            },
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            model_update_frequency="5 days",
            training_as_of_date_frequencies=["1 days"],
            test_as_of_date_frequencies=["1 days"],
            max_training_histories=["5 days"],
            test_durations=["5 days"],
            test_label_timespans=["1 day"],
            training_label_timespans=["1 day"],
        )
        result = chopper.chop_time()
        assert result == expected_result

    def test_training_label_timespan_longer_than_1_day(self):
        expected_result = [
            {
                "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
                "label_start_time": datetime.datetime(2010, 1, 1, 0, 0),
                "feature_end_time": datetime.datetime(2010, 1, 19, 0, 0),
                "label_end_time": datetime.datetime(2010, 1, 19, 0, 0),
                "train_matrix": {
                    "first_as_of_time": datetime.datetime(2010, 1, 1, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 4, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 9, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 1, 0, 0),
                        datetime.datetime(2010, 1, 2, 0, 0),
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0),
                    ],
                    "training_label_timespan": "5 days",
                    "training_as_of_date_frequency": "1 days",
                    "max_training_history": "5 days",
                },
                "test_matrices": [
                    {
                        "first_as_of_time": datetime.datetime(2010, 1, 9, 0, 0),
                        "last_as_of_time": datetime.datetime(2010, 1, 13, 0, 0),
                        "matrix_info_end_time": datetime.datetime(2010, 1, 18, 0, 0),
                        "as_of_times": [
                            datetime.datetime(2010, 1, 9, 0, 0),
                            datetime.datetime(2010, 1, 10, 0, 0),
                            datetime.datetime(2010, 1, 11, 0, 0),
                            datetime.datetime(2010, 1, 12, 0, 0),
                            datetime.datetime(2010, 1, 13, 0, 0),
                        ],
                        "test_label_timespan": "5 days",
                        "test_as_of_date_frequency": "1 days",
                        "test_duration": "5 days",
                    }
                ],
            }
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 19, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 19, 0, 0),
            model_update_frequency="5 days",
            training_as_of_date_frequencies=["1 days"],
            test_as_of_date_frequencies=["1 days"],
            max_training_histories=["5 days"],
            test_durations=["5 days"],
            test_label_timespans=["5 days"],
            training_label_timespans=["5 days"],
        )
        result = chopper.chop_time()
        assert result == expected_result

    def test_unevenly_divisible_lookback_duration(self):
        expected_result = [
            {
                "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
                "label_start_time": datetime.datetime(2010, 1, 1, 0, 0),
                "feature_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "label_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "train_matrix": {
                    "first_as_of_time": datetime.datetime(2010, 1, 1, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 4, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 5, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 1, 0, 0),
                        datetime.datetime(2010, 1, 2, 0, 0),
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0),
                    ],
                    "training_label_timespan": "1 day",
                    "training_as_of_date_frequency": "1 days",
                    "max_training_history": "7 days",
                },
                "test_matrices": [
                    {
                        "first_as_of_time": datetime.datetime(2010, 1, 5, 0, 0),
                        "last_as_of_time": datetime.datetime(2010, 1, 9, 0, 0),
                        "matrix_info_end_time": datetime.datetime(2010, 1, 10, 0, 0),
                        "as_of_times": [
                            datetime.datetime(2010, 1, 5, 0, 0),
                            datetime.datetime(2010, 1, 6, 0, 0),
                            datetime.datetime(2010, 1, 7, 0, 0),
                            datetime.datetime(2010, 1, 8, 0, 0),
                            datetime.datetime(2010, 1, 9, 0, 0),
                        ],
                        "test_label_timespan": "1 day",
                        "test_as_of_date_frequency": "1 days",
                        "test_duration": "5 days",
                    }
                ],
            },
            {
                "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
                "label_start_time": datetime.datetime(2010, 1, 1, 0, 0),
                "feature_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "label_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "train_matrix": {
                    "first_as_of_time": datetime.datetime(2010, 1, 2, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 9, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 10, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 2, 0, 0),
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0),
                        datetime.datetime(2010, 1, 5, 0, 0),
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0),
                    ],
                    "training_label_timespan": "1 day",
                    "training_as_of_date_frequency": "1 days",
                    "max_training_history": "7 days",
                },
                "test_matrices": [
                    {
                        "first_as_of_time": datetime.datetime(2010, 1, 10, 0, 0),
                        "last_as_of_time": datetime.datetime(2010, 1, 14, 0, 0),
                        "matrix_info_end_time": datetime.datetime(2010, 1, 15, 0, 0),
                        "as_of_times": [
                            datetime.datetime(2010, 1, 10, 0, 0),
                            datetime.datetime(2010, 1, 11, 0, 0),
                            datetime.datetime(2010, 1, 12, 0, 0),
                            datetime.datetime(2010, 1, 13, 0, 0),
                            datetime.datetime(2010, 1, 14, 0, 0),
                        ],
                        "test_label_timespan": "1 day",
                        "test_as_of_date_frequency": "1 days",
                        "test_duration": "5 days",
                    }
                ],
            },
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            model_update_frequency="5 days",
            training_as_of_date_frequencies=["1 days"],
            test_as_of_date_frequencies=["1 days"],
            max_training_histories=["7 days"],
            test_durations=["5 days"],
            test_label_timespans=["1 day"],
            training_label_timespans=["1 day"],
        )
        result = chopper.chop_time()
        assert result == expected_result

    def test_unevenly_divisible_update_window(self):
        expected_result = [
            {
                "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
                "label_start_time": datetime.datetime(2010, 1, 3, 0, 0),
                "feature_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "label_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "train_matrix": {
                    "first_as_of_time": datetime.datetime(2010, 1, 3, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 4, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 5, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0),
                    ],
                    "training_label_timespan": "1 day",
                    "training_as_of_date_frequency": "1 days",
                    "max_training_history": "5 days",
                },
                "test_matrices": [
                    {
                        "first_as_of_time": datetime.datetime(2010, 1, 5, 0, 0),
                        "last_as_of_time": datetime.datetime(2010, 1, 9, 0, 0),
                        "matrix_info_end_time": datetime.datetime(2010, 1, 10, 0, 0),
                        "as_of_times": [
                            datetime.datetime(2010, 1, 5, 0, 0),
                            datetime.datetime(2010, 1, 6, 0, 0),
                            datetime.datetime(2010, 1, 7, 0, 0),
                            datetime.datetime(2010, 1, 8, 0, 0),
                            datetime.datetime(2010, 1, 9, 0, 0),
                        ],
                        "test_label_timespan": "1 day",
                        "test_as_of_date_frequency": "1 days",
                        "test_duration": "5 days",
                    }
                ],
            },
            {
                "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
                "label_start_time": datetime.datetime(2010, 1, 3, 0, 0),
                "feature_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "label_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                "train_matrix": {
                    "first_as_of_time": datetime.datetime(2010, 1, 4, 0, 0),
                    "last_as_of_time": datetime.datetime(2010, 1, 9, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 10, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 4, 0, 0),
                        datetime.datetime(2010, 1, 5, 0, 0),
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0),
                    ],
                    "training_label_timespan": "1 day",
                    "training_as_of_date_frequency": "1 days",
                    "max_training_history": "5 days",
                },
                "test_matrices": [
                    {
                        "first_as_of_time": datetime.datetime(2010, 1, 10, 0, 0),
                        "last_as_of_time": datetime.datetime(2010, 1, 14, 0, 0),
                        "matrix_info_end_time": datetime.datetime(2010, 1, 15, 0, 0),
                        "as_of_times": [
                            datetime.datetime(2010, 1, 10, 0, 0),
                            datetime.datetime(2010, 1, 11, 0, 0),
                            datetime.datetime(2010, 1, 12, 0, 0),
                            datetime.datetime(2010, 1, 13, 0, 0),
                            datetime.datetime(2010, 1, 14, 0, 0),
                        ],
                        "test_label_timespan": "1 day",
                        "test_as_of_date_frequency": "1 days",
                        "test_duration": "5 days",
                    }
                ],
            },
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 3, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            model_update_frequency="5 days",
            training_as_of_date_frequencies=["1 days"],
            test_as_of_date_frequencies=["1 days"],
            max_training_histories=["5 days"],
            test_durations=["5 days"],
            test_label_timespans=["1 day"],
            training_label_timespans=["1 day"],
        )
        result = chopper.chop_time()
        assert result == expected_result


class test__init__(TestCase):
    def test_bad_feature_start_time(self):
        with self.assertRaises(ValueError):
            Timechop(
                feature_start_time=datetime.datetime(2011, 1, 1, 0, 0),
                feature_end_time=datetime.datetime(2010, 1, 16, 0, 0),
                label_start_time=datetime.datetime(2010, 1, 3, 0, 0),
                label_end_time=datetime.datetime(2010, 1, 16, 0, 0),
                model_update_frequency="5 days",
                training_as_of_date_frequencies=["1 days"],
                test_as_of_date_frequencies=["1 days"],
                max_training_histories=["5 days"],
                test_durations=["5 days"],
                test_label_timespans=["1 day"],
                training_label_timespans=["1 day"],
            )



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import os
import yaml

def fill_timechop_config_missing(config, db_engine):
    """
    Fill with default values the temporal_config params if they are missing

    Args:
        config (dict) a triage experiment configuration
        db_engine (psycopg2 connection) a db connection

    Returns: (dict) a triage temporal_config
    """
    timechop_config = config['temporal_config']

    default_config = {'model_update_frequency': '100y',
                      'training_as_of_date_frequencies': '100y',
                      'test_as_of_date_frequencies': '100y',
                      'max_training_histories': '0d',
                      'test_durations': '0d',
                      }

    # Checks if label_timespan is present
    if 'label_timespans' in timechop_config.keys():
        if any([k in timechop_config.keys() for k in ['training_label_timespans', 'test_labels_timespans']]):
            raise KeyError("You can't always get what you want, but just sometimes, you get what you need: The config file has conflicting keys: the 'label_timespan' and 'training_label_timespans' and/or 'test_label_timespans'")
        default_config['training_label_timespans'] = default_config['test_label_timespans'] = timechop_config['label_timespans']
        timechop_config.pop('label_timespans') ## We don't need this value anymore

    # Checks if some of the date range  limits  is missing, if so replaces with
    # min, max accordingy from de from_objs
    if any([k not in timechop_config.keys() for k in ['feature_start_time', 'feature_end_time', 'label_start_time', 'label_end_time']]):
        from_query = "(select min({knowledge_date}) as min_date, max({knowledge_date}) as max_date from (select * from {from_obj}) as t)"

        feature_aggregations = config['feature_aggregations']

        from_queries = [from_query.format(knowledge_date = agg['knowledge_date_column'], from_obj=agg['from_obj']) for agg in feature_aggregations]

        unions = "\n union \n".join(from_queries)

        query = "select to_char(min(min_date), 'YYYY-MM-DD'), to_char(max(max_date), 'YYYY-MM-DD') from ({unions}) as u".format(unions=unions)

        with db_engine.connect() as conn:
            rs = conn.execute(query)
            min_date, max_date = rs.fetchall()[0]

        default_config['feature_start_time'] = default_config['label_start_time'] = min_date
        default_config['feature_end_time'] = default_config['label_end_time'] = max_date

    # Replaces missing values
    default_config.update(timechop_config)

    return default_config


def fill_cohort_config_missing(config):
    """
    If none cohort_config section is provided, include all the entities by default

    Args:
        config (dict) a triage experiment configuration

    Returns: (dict) a triage cohort config
    """
    from_query = "(select entity_id, {knowledge_date} as knowledge_date from (select * from {from_obj}) as t)"

    feature_aggregations = config['feature_aggregations']

    from_queries = [from_query.format(knowledge_date = agg['knowledge_date_column'], from_obj=agg['from_obj']) for agg in feature_aggregations]

    unions = "\n union \n".join(from_queries)

    query = f"select distinct entity_id from ({unions}) as e" +" where knowledge_date < '{as_of_date}'"

    cohort_config = config.get('cohort_config', {})
    default_config = {'query': query, 'name': 'all_entities'}

    default_config.update(cohort_config)

    return default_config


def fill_feature_group_definition(config):
    """
    If feature_group_definition is not presents, this function sets it to all
    the distinct feature_aggregations' prefixes

    Args:
        config (dict) a triage experiment configuration

    Returns: (dict) a triage feature_group config
    """
    feature_group_definition = config.get('feature_group_definition', {})
    if not feature_group_definition:
        feature_aggregations = config['feature_aggregations']

        feature_group_definition['prefix'] = list({agg['prefix'] for agg in feature_aggregations})

    return feature_group_definition


def fill_model_grid_presets(config):
    """Determine if model grid preset is being used and return the appropriate grid if so

       Args:
            config (dict) a triage experiment configuration

        Returns: (dict) a triage model grid config
    """

    grid_config = config.get('grid_config')
    preset_type = config.get('model_grid_preset')

    if preset_type is not None:
        grid_config = model_grid_preset(preset_type, grid_config)

    return grid_config


def model_grid_preset(grid_type, grid_config=None):
    """Load a preset model grid.

       Args:
            grid_type (string) The type of preset grid to load. May
                by `quickstart`, `small`, `medium`, `large`, or `texas`
            grid_config (dict) The user-specified model grid, allowing
                users to extend a preset grid with other models, such
                as common-sense baselines specific to their project

        Returns: (dict) a triage model grid config
    """

    presets_file = os.path.join(os.path.dirname(__file__), 'model_grid_presets.yaml')
    with open(presets_file, 'r') as f:
        model_grid_presets = yaml.full_load(f)

    # output is a collector for the resulting grid, so initialize with the user-specified
    # triage grid (if present), otherwise start with an empty dict. We initialize
    # prev_type with the preset grid type to start crawling the presets at that point
    output = (grid_config or {}).copy()
    prev_type = grid_type

    # collapse the grid parameters down the levels until we reach one with no lower level
    while prev_type is not None:
        prev = model_grid_presets[prev_type]['grid'].copy()

        # look for new model types and hyperparameters to incorporate into the output
        for model_type in set(output.keys()).union(set(prev.keys())):
            curr_model = output.get(model_type, {}).copy()
            # if the model type exists in the lower-level preset, update any associated hyperparameter
            # values in the output (those only in the higher level grid will pass through unchanged)
            for hyperparam in prev.get(model_type, {}).keys():
                curr_model[hyperparam] = sorted(list(set(curr_model.get(hyperparam, []) + prev[model_type][hyperparam])), key=lambda x: x if x is not None else 0)
            output[model_type] = curr_model

        # traverse the linked list to one level deeper and repeat
        prev_type = model_grid_presets[prev_type]['prev']

    return output



"""Classes representing simple but deep data structures that we reuse throughout
Triage code and want to display more intelligently in log files
"""


class TruncatedRepresentationList(list):
    def __repr__(self):
        total = len(self)
        if total != 1:
            return f"[{self[0]} ... {self[-1]} (Total: {total})]"
        else:
            return f"[{self[0]}] (Total: {total})"


class AsOfTimeList(TruncatedRepresentationList):
    pass


class FeatureNameList(TruncatedRepresentationList):
    pass



from datetime import date, timedelta

import testing.postgresql
from sqlalchemy import create_engine

from triage.component.architect.label_generators import LabelGenerator

from .utils import create_binary_outcome_events


# Sample events data to use for all tests
events_data = [
    # entity id, event_date, outcome
    [1, date(2014, 1, 1), True],
    [1, date(2014, 11, 10), False],
    [1, date(2015, 1, 1), False],
    [1, date(2015, 11, 10), True],
    [2, date(2014, 6, 8), True],
    [2, date(2015, 6, 8), False],
    [3, date(2014, 3, 3), False],
    [3, date(2014, 7, 24), False],
    [3, date(2015, 3, 3), True],
    [3, date(2015, 7, 24), False],
    [4, date(2014, 12, 13), False],
    [4, date(2015, 12, 13), False],
]


# An example label generation query to use for all tests.
# Since this is expected to be passed in properly by the user
# we don't need to test variations. Just reuse this one.
LABEL_GENERATE_QUERY = """select
    events.entity_id,
    bool_or(outcome::bool)::integer as outcome
from events
where
    '{as_of_date}' <= outcome_date
    and outcome_date < '{as_of_date}'::timestamp + interval '{label_timespan}'
    group by entity_id
"""

LABELS_TABLE_NAME = "labels"


def test_label_generation():
    # Generate labels for one as-of-date/label timespan combo
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_binary_outcome_events(engine, "events", events_data)

        label_generator = LabelGenerator(db_engine=engine, query=LABEL_GENERATE_QUERY)
        label_generator._create_labels_table(LABELS_TABLE_NAME)
        label_generator.generate(
            start_date="2014-09-30", label_timespan="6months", labels_table="labels"
        )

        expected = [
            # entity_id, as_of_date, label_timespan, name, type, label
            (1, date(2014, 9, 30), timedelta(180), "outcome", "binary", False),
            (3, date(2014, 9, 30), timedelta(180), "outcome", "binary", True),
            (4, date(2014, 9, 30), timedelta(180), "outcome", "binary", False),
        ]
        result = engine.execute(
            "select * from {} order by entity_id, as_of_date".format(LABELS_TABLE_NAME)
        )
        records = [row for row in result]
        assert records == expected


def test_generate_all_labels_replace():
    # Generate labels for combinations of as-of-date and label timespan
    # use replace=True
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_binary_outcome_events(engine, "events", events_data)

        label_generator = LabelGenerator(db_engine=engine, query=LABEL_GENERATE_QUERY, replace=True)
        label_generator.generate_all_labels(
            labels_table=LABELS_TABLE_NAME,
            as_of_dates=["2014-09-30", "2015-03-30"],
            label_timespans=["6month", "3month"],
        )

        result = engine.execute(
            """
            select * from {}
            order by entity_id, as_of_date, label_timespan desc
        """.format(
                LABELS_TABLE_NAME
            )
        )
        records = [row for row in result]

        expected = [
            # entity_id, as_of_date, label_timespan, name, type, label
            (1, date(2014, 9, 30), timedelta(180), "outcome", "binary", False),
            (1, date(2014, 9, 30), timedelta(90), "outcome", "binary", False),
            (2, date(2015, 3, 30), timedelta(180), "outcome", "binary", False),
            (2, date(2015, 3, 30), timedelta(90), "outcome", "binary", False),
            (3, date(2014, 9, 30), timedelta(180), "outcome", "binary", True),
            (3, date(2015, 3, 30), timedelta(180), "outcome", "binary", False),
            (4, date(2014, 9, 30), timedelta(180), "outcome", "binary", False),
            (4, date(2014, 9, 30), timedelta(90), "outcome", "binary", False),
        ]
        assert records == expected


def test_generate_all_labels_noreplace():
    # test the 'replace=False' functionality
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_binary_outcome_events(engine, "events", events_data)

        label_generator = LabelGenerator(
            db_engine=engine,
            query=LABEL_GENERATE_QUERY,
            replace=False
        )
        label_generator.generate_all_labels(
            labels_table=LABELS_TABLE_NAME,
            as_of_dates=["2014-09-30"],
            label_timespans=["6month"],
        )

        result = engine.execute(
            """
            select * from {}
            order by entity_id, as_of_date, label_timespan desc
        """.format(
                LABELS_TABLE_NAME
            )
        )
        records = [row for row in result]

        expected = [
            # entity_id, as_of_date, label_timespan, name, type, label
            (1, date(2014, 9, 30), timedelta(180), "outcome", "binary", False),
            (3, date(2014, 9, 30), timedelta(180), "outcome", "binary", True),
            (4, date(2014, 9, 30), timedelta(180), "outcome", "binary", False),
        ]
        assert records == expected

        # round 2
        label_generator.generate_all_labels(
            labels_table=LABELS_TABLE_NAME,
            as_of_dates=["2014-09-30", "2015-03-30"],
            label_timespans=["6month", "3month"],
        )

        result = engine.execute(
            """
            select * from {}
            order by entity_id, as_of_date, label_timespan desc
        """.format(
                LABELS_TABLE_NAME
            )
        )
        records = [row for row in result]

        expected = [
            # entity_id, as_of_date, label_timespan, name, type, label
            (1, date(2014, 9, 30), timedelta(180), "outcome", "binary", False),
            (1, date(2014, 9, 30), timedelta(90), "outcome", "binary", False),
            (2, date(2015, 3, 30), timedelta(180), "outcome", "binary", False),
            (2, date(2015, 3, 30), timedelta(90), "outcome", "binary", False),
            (3, date(2014, 9, 30), timedelta(180), "outcome", "binary", True),
            (3, date(2015, 3, 30), timedelta(180), "outcome", "binary", False),
            (4, date(2014, 9, 30), timedelta(180), "outcome", "binary", False),
            (4, date(2014, 9, 30), timedelta(90), "outcome", "binary", False),
        ]
        assert records == expected



from itertools import product
import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from .selection_rules import BoundSelectionRule
from .utils import make_list


def _expand_param_sets(rule_instances, values):
    expanded_param_sets = []
    for v in product(*values):
        params = dict(zip(rule_instances, v))
        expanded_param_sets.append(params)
    return expanded_param_sets


def _build_rule_arguments(expanded_param_set, shared_param_set):
    rule_args = {}
    for key, value in expanded_param_set.items():
        rule_args[key] = value
    for key, value in shared_param_set.items():
        rule_args[key] = value
    return rule_args


def _bound_rules_from(shared_param_set, selection_rule):
    rules = []
    rule_instances_and_values = [
        (key, make_list(value)) for key, value in selection_rule.items()
    ]
    rule_instances, values = zip(*rule_instances_and_values)

    for expanded_param_set in _expand_param_sets(rule_instances, values):
        function_name = expanded_param_set["name"]
        del expanded_param_set["name"]
        rules.append(
            BoundSelectionRule(
                function_name=function_name,
                args=_build_rule_arguments(expanded_param_set, shared_param_set),
            )
        )
    return rules


def make_selection_rule_grid(rule_groups):
    """Convert a compact selection rule group representation to a
        list of bound selection rules.

    Arguments:
        rule_groups (list): List of dicts used to specify selection rule grid. 
        
    Most users will want to use [rulemaker objects](#rulemakers)
    to generate their `rule_group` specifications.
    
    An example rule_groups specification:

    ```
    [{
            'shared_parameters': [
                    {'metric': 'precision@', 'parameter': '100_abs'},
                    {'metric': 'recall@', 'parameter': '100_abs'},
                ],
                'selection_rules': [
                    {'name': 'most_frequent_best_dist', 'dist_from_best_case': [0.1, 0.2, 0.3]},
                    {'name': 'best_current_value'}
                ]
        }, {
            'shared_parameters': [
                {'metric1': 'precision@', 'parameter1': '100_abs'},
            ],
            'selection_rules': [
                {
                    'name': 'best_average_two_metrics',
                    'metric2': ['recall@'],
                    'parameter2': ['100_abs'],
                    'metric1_weight': [0.4, 0.5, 0.6]
                },
            ]
        }]
    ```
    Returns:
        list: list of audition.selection_rules.BoundSelectionRule objects"""


    rules = []
    logger.debug("Expanding selection rule groups into full grid")
    for rule_group in rule_groups:
        logger.debug("Expanding rule group %s", rule_group)
        for shared_param_set, selection_rule in product(
            rule_group["shared_parameters"], rule_group["selection_rules"]
        ):
            logger.debug(
                "Expanding shared param set %s and selection rules %s",
                shared_param_set,
                selection_rule,
            )
            new_rules = _bound_rules_from(shared_param_set, selection_rule)
            logger.debug("Found %s new rules", len(new_rules))
            rules += new_rules
    logger.debug(
        "Found %s total selection rules. Full list: %s",
        len(rules),
        [rule.descriptive_name for rule in rules],
    )
    return rules



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from sqlalchemy.orm import sessionmaker

from triage.component.architect.entity_date_table_generators import EntityDateTableGenerator
from triage.component.catwalk.utils import (filename_friendly_hash, get_subset_table_name)
from triage.component.results_schema import Subset


class SubsetterNoOp:
    def generate_tasks(self, subset_configs):
        logger.debug(
            "No subsets configuration is available, so subsets tasks will not be created"
        )

        return []

    def process_all_tasks(self, tasks):
        logger.notice(
            "No subsets configuration is available, so subsets will not be created"
        )

    def process_task(self, subset_config, subset_hash, subset_table_generator):
        logger.notice(
            "No subsets configuration is available, so subset task  will not be created"
        )

    def save_subset_to_db(self, subset_hash, subset_config):
        logger.notice(
            "No subsets configuration is available, so subsets will not be created"
        )

class Subsetter:
    def __init__(
        self,
        db_engine,
        replace,
        as_of_times,
    ):
        self.db_engine = db_engine
        self.replace = replace
        self.as_of_times = as_of_times

    def generate_tasks(self, subset_configs):
        logger.debug("Generating subset table creation tasks")
        subset_tasks = []
        for subset_config in subset_configs:
            if subset_config:
                subset_hash = filename_friendly_hash(subset_config)
                subset_table_generator = EntityDateTableGenerator(
                    entity_date_table_name=get_subset_table_name(subset_config),
                    db_engine=self.db_engine,
                    query=subset_config["query"],
                    replace=self.replace
                )
                subset_tasks.append(
                    {
                        "subset_config": subset_config,
                        "subset_hash": subset_hash,
                        "subset_table_generator": subset_table_generator,
                    }
                )
        return subset_tasks

    def process_all_tasks(self, tasks):
        logger.info("Creating subsets")
        for task in tasks:
            self.process_task(**task)
        logger.success("Subsets stored successfully")

    def process_task(self, subset_config, subset_hash, subset_table_generator):
        logger.debug(
            f"Creating subset for {subset_config['name']}-{subset_hash}"
        )
        subset_table_generator.generate_entity_date_table(
            as_of_dates=self.as_of_times
        )
        self.save_subset_to_db(subset_hash, subset_config)
        logger.debug(f"Subset {subset_config['name']}-{subset_hash} created successfully")

    def save_subset_to_db(self, subset_hash, subset_config):
        session = sessionmaker(bind=self.db_engine)()
        session.merge(Subset(subset_hash=subset_hash, config=subset_config))
        session.commit()
        session.close()



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import re

from dateutil.relativedelta import relativedelta
from datetime import datetime



def parse_from_obj(config, alias):
    """
    Parses a from_obj configuration key. If it's a from_obj_table just returns it.
    If it's a from_obj_query creates the sub_query with alias
    Args:
        config: the yaml dict
        alias: the name of the alias if there's a from_obj_query

    Returns:

    """
    from_obj = config.get("from_obj_table", None)
    if not from_obj:
        from_obj = config.get("from_obj_query", None)
        return " ({}) {} ".format(from_obj, alias) if from_obj else None
    return from_obj

def dt_from_str(dt_str):
    if isinstance(dt_str, datetime):
        return dt_str
    return datetime.strptime(dt_str, "%Y-%m-%d")


def parse_delta_string(delta_string):
    """Given a string in a postgres interval format (e.g., '1 month'),
    parse the units and value from it.

    Assumptions:
    - The string is in the format 'value unit', where
      value is an int and unit is one of year(s), month(s), day(s),
      week(s), hour(s), minute(s), second(s), microsecond(s), or an
      abbreviation matching y, d, w, h, m, s, or ms (case-insensitive).
      For example: 1 year, 1year, 2 years, 1 y, 2y, 1Y.

    :param delta_string: the time interval to convert
    :type delta_string: str

    :return: time units, number of units (value)
    :rtype: tuple

    :raises: ValueError if the delta_string is not in the expected format

    """
    match = parse_delta_string.pattern.search(delta_string)
    if match:
        (pre_value, units) = match.groups()
        return (units, int(pre_value))

    raise ValueError(
        "Could not parse value from time delta string: {!r}".format(delta_string)
    )


parse_delta_string.pattern = re.compile(r"^(\d+) *([^ ]+)$")


def convert_str_to_relativedelta(delta_string):
    """Given a string in a postgres interval format (e.g., '1 month'),
    convert it to a dateutil.relativedelta.relativedelta.

    Assumptions:
    - The string is in the format 'value unit', where
      value is an int and unit is one of year(s), month(s), day(s),
      week(s), hour(s), minute(s), second(s), microsecond(s), or an
      abbreviation matching y, d, w, h, m, s, or ms (case-insensitive).
      For example: 1 year, 1year, 2 years, 1 y, 2y, 1Y.

    :param delta_string: the time interval to convert
    :type delta_string: str

    :return: the time interval as a relativedelta
    :rtype: dateutil.relativedelta.relativedelta

    :raises: ValueError if the delta_string is not in the expected format

    """
    (units, value) = parse_delta_string(delta_string)

    verbose_match = convert_str_to_relativedelta.pattern_verbose.search(units)
    if verbose_match:
        unit_type = verbose_match.group(1) + "s"
        return relativedelta(**{unit_type: value})

    try:
        unit_type = convert_str_to_relativedelta.brief_units[units.lower()]
    except KeyError:
        pass
    else:
        if unit_type == "minutes":
            logger.warning(
                f'Time delta units "{units}" converted to minutes.'
            )
        return relativedelta(**{unit_type: value})

    raise ValueError("Could not handle units. Units: {} Value: {}".format(units, value))


convert_str_to_relativedelta.pattern_verbose = re.compile(
    r"^(year|month|day|week|hour|minute|second|microsecond)s?$"
)

convert_str_to_relativedelta.brief_units = {
    "y": "years",
    "d": "days",
    "w": "weeks",
    "h": "hours",
    "m": "minutes",
    "s": "seconds",
    "ms": "microseconds",
}



from triage.component.catwalk.utils import (
    filename_friendly_hash,
    save_experiment_and_get_hash,
    associate_models_with_experiment,
    associate_matrices_with_experiment,
    missing_model_hashes,
    missing_matrix_uuids,
    sort_predictions_and_labels,
)
from triage.component.results_schema.schema import Matrix, Model
from triage.component.catwalk.db import ensure_db
from sqlalchemy import create_engine
import testing.postgresql
import datetime
import re
import numpy as np
from numpy.testing import assert_array_equal
import pytest


def test_filename_friendly_hash():
    data = {
        "stuff": "stuff",
        "other_stuff": "more_stuff",
        "a_datetime": datetime.datetime(2015, 1, 1),
        "a_date": datetime.date(2016, 1, 1),
        "a_number": 5.0,
    }
    output = filename_friendly_hash(data)
    assert isinstance(output, str)
    assert re.match(r"^[\w]+$", output) is not None

    # make sure ordering keys differently doesn't change the hash
    new_output = filename_friendly_hash(
        {
            "other_stuff": "more_stuff",
            "stuff": "stuff",
            "a_datetime": datetime.datetime(2015, 1, 1),
            "a_date": datetime.date(2016, 1, 1),
            "a_number": 5.0,
        }
    )
    assert new_output == output

    # make sure new data hashes to something different
    new_output = filename_friendly_hash({"stuff": "stuff", "a_number": 5.0})
    assert new_output != output


def test_filename_friendly_hash_stability():
    nested_data = {"one": "two", "three": {"four": "five", "six": "seven"}}
    output = filename_friendly_hash(nested_data)
    # 1. we want to make sure this is stable across different runs
    # so hardcode an expected value
    assert output == "9a844a7ebbfd821010b1c2c13f7391e6"
    other_nested_data = {"one": "two", "three": {"six": "seven", "four": "five"}}
    new_output = filename_friendly_hash(other_nested_data)
    assert output == new_output


def test_save_experiment_and_get_hash():
    # no reason to make assertions on the config itself, use a basic dict
    experiment_config = {"one": "two"}
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ensure_db(engine)
        exp_hash = save_experiment_and_get_hash(experiment_config, 1234, engine)
        assert isinstance(exp_hash, str)
        new_hash = save_experiment_and_get_hash(experiment_config, 1234, engine)
        assert new_hash == exp_hash


def test_missing_model_hashes():
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)

        experiment_hash = save_experiment_and_get_hash({}, 1234, db_engine)
        model_hashes = ['abcd', 'bcde', 'cdef']

        # if we associate model hashes with an experiment but don't actually train the models
        # they should show up as missing
        associate_models_with_experiment(experiment_hash, model_hashes, db_engine)
        assert missing_model_hashes(experiment_hash, db_engine) == model_hashes

        # if we insert a model row they should no longer be considered missing
        db_engine.execute(
            f"insert into {Model.__table__.fullname} (model_hash) values (%s)",
            model_hashes[0]
        )
        assert missing_model_hashes(experiment_hash, db_engine) == model_hashes[1:]


def test_missing_matrix_uuids():
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)

        experiment_hash = save_experiment_and_get_hash({}, 1234, db_engine)
        matrix_uuids = ['abcd', 'bcde', 'cdef']

        # if we associate matrix uuids with an experiment but don't actually build the matrices
        # they should show up as missing
        associate_matrices_with_experiment(experiment_hash, matrix_uuids, db_engine)
        assert missing_matrix_uuids(experiment_hash, db_engine) == matrix_uuids

        # if we insert a matrix row they should no longer be considered missing
        db_engine.execute(
            f"insert into {Matrix.__table__.fullname} (matrix_uuid) values (%s)",
            matrix_uuids[0]
        )
        assert missing_matrix_uuids(experiment_hash, db_engine) == matrix_uuids[1:]


def test_sort_predictions_and_labels():
    predictions = np.array([0.5, 0.4, 0.6, 0.5, 0.6])
    entities = np.array(range(5))
    labels = np.array([0, 0, 1, 1, None])

    # best sort
    sorted_predictions, sorted_labels, sorted_entities = sort_predictions_and_labels(
        predictions, labels, entities, tiebreaker='best'
    )
    assert_array_equal(sorted_predictions, np.array([0.6, 0.6, 0.5, 0.5, 0.4]))
    assert_array_equal(sorted_labels, np.array([1, None, 1, 0, 0]))
    assert_array_equal(sorted_entities.to_numpy(), np.array([2, 4, 3, 0, 1]))

    # worst sort
    sorted_predictions, sorted_labels, sorted_entities = sort_predictions_and_labels(
        predictions, labels, entities, tiebreaker='worst'
    )
    assert_array_equal(sorted_predictions, np.array([0.6, 0.6, 0.5, 0.5, 0.4]))
    assert_array_equal(sorted_labels, np.array([None, 1, 0, 1, 0]))
    assert_array_equal(sorted_entities.to_numpy(), np.array([4, 2, 0, 3, 1]))

    # random tiebreaker needs a seed
    with pytest.raises(ValueError):
        sort_predictions_and_labels(predictions, labels, entities, tiebreaker='random')

    # random tiebreaker respects the seed
    sorted_predictions, sorted_labels, sorted_entities = sort_predictions_and_labels(
        predictions,
        labels,
        entities,
        tiebreaker='random',
        sort_seed=1234
    )
    assert_array_equal(sorted_predictions, np.array([0.6, 0.6, 0.5, 0.5, 0.4]))
    assert_array_equal(sorted_labels, np.array([None, 1, 1, 0, 0]))
    assert_array_equal(sorted_entities.to_numpy(), np.array([4, 2, 3, 0, 1]))

    sorted_predictions, sorted_labels, sorted_entities = sort_predictions_and_labels(
        predictions,
        labels,
        entities,
        tiebreaker='random',
        sort_seed=24376234
    )
    assert_array_equal(sorted_predictions, np.array([0.6, 0.6, 0.5, 0.5, 0.4]))
    assert_array_equal(sorted_labels, np.array([None, 1, 0, 1, 0]))
    assert_array_equal(sorted_entities.to_numpy(), np.array([4, 2, 0, 3, 1]))



from scipy import stats
import numpy as np
from triage.component.catwalk.exceptions import BaselineFeatureNotInMatrix


class PercentileRankOneFeature:
    def __init__(self, feature, descend=False):
        self.feature = feature  # which feature to rank on
        self.descend = (
            descend
        )  # should feature be ranked so lower values -> higher scores
        self.feature_importances_ = None

    def _set_feature_importances_(self, x):
        """ Assigns feature importances following the rule: 1 for the feature we
        are ranking on, 0 for all other features.
        """
        feature_importances = [0] * len(x.columns)
        try:
            position = x.columns.get_loc(self.feature)
        except KeyError:
            raise BaselineFeatureNotInMatrix(
                (
                    "Trying to rank on a feature ({feature_name})"
                    " not included in the training matrix!".format(
                        feature_name=self.feature
                    )
                )
            )
        feature_importances[position] = 1
        self.feature_importances_ = np.array(feature_importances)

    def fit(self, x, y):
        """ Set feature importances and return self.
        """
        self._set_feature_importances_(x)
        return self

    def predict_proba(self, x):
        """ Generate the rank percentile scores and return these.
        """
        # reduce x to the selected feature, raise error if not found
        x = x[self.feature]

        # we need different behavior depending on rank ordering. percentiles
        # should be able to be interpreted as "proportion of entities ranking
        # BELOW this entity's value". scipy will assign lower ranks to lower
        # values of the feature. so if the entities have values [0, 0, 1, 2, 2],
        # the first two entities will have the lowest ranks (and therefore the
        # lowest risk scores) and the last two will have the highest ranks (and
        # highest risk scores). for the descending method, we need to reverse
        # this, and for both sorting directions, we need to convert the ranks to
        # percentiles.

        # when ascending: tied entities should get the *lowest* rank, so for
        # [0, 0, 1, 2, 2] the ranks should be [1, 1, 3, 4, 4]. these can be
        # converted to the number of entities below each value by subtracting 1
        # from each rank, yielding [0, 0, 2, 3, 3]. from here, we can calculate
        # the proportions by dividing by the length of each list.
        method = "min"
        subtract = 1

        # when descending: tied entities should get the *highest* rank, so for
        # [0, 0, 1, 2, 2] the ranks should be [2, 2, 3, 5, 5]. if we reverse
        # these ranks by substracting all items from the maximum rank (5), we
        # end up with the correct ranks for calculating percentiles:
        # [3, 3, 2, 0, 0]. to simplify the code, we first divide by the length
        # of the list then subtract the result from the maxmimum percentile (1).
        # it produces the same result as subtracting from 5 then dividing:
        #   ([5, 5, 5, 5, 5] -  [2, 2, 3, 5, 5]) / 5  = [0.6, 0.6, 0.4, 0, 0]
        # and
        #    [1, 1, 1, 1, 1] - ([2, 2, 3, 5, 5]  / 5) = [0.6, 0.6, 0.4, 0, 0]
        if self.descend:
            method = "max"
            subtract = 0

        # get the ranks and convert to percentiles
        ranks = stats.rankdata(x, method)
        ranks = [(rank - subtract) / len(x) for rank in ranks]
        if self.descend:
            ranks = [1 - rank for rank in ranks]

        # format it like sklearn output and return
        return np.array([np.zeros(len(x)), ranks]).transpose()



import matplotlib
import numpy as np
import matplotlib.lines as mlines

matplotlib.use("Agg")
from matplotlib import pyplot as plt  # noqa: E402


def plot_bounds(observed_min, observed_max):
    """Compute the plot bounds for observed data

    Args:
        observed_min (number) -- Lowest number found in data
        observed_max (number) -- Highest number found in data

    Returns: (number, number) A minimum and maximum x value for the plot
    """
    if observed_min >= 0.0 and observed_max <= 1.0:
        plot_min, plot_max = (0.0, 1.0)
    else:
        # 10% padding on the high end
        padding = 0.1 * (observed_max - float(observed_min))
        plot_min = observed_min
        plot_max = observed_max + padding

    return plot_min, plot_max


def category_colordict(cmap_name, categories, highlight_grp=None):
    # want to step through the discrete color map rather than sampling
    # across the entire range, so create an even spacing from 0 to 1
    # with as many steps as in the color map (cmap.N), then repeat it
    # enough times to ensure we cover all our categories
    cmap = plt.get_cmap(cmap_name)
    categories_with_colors = [cat for cat in categories if cat != highlight_grp]
    ncyc = int(np.ceil(1.0 * len(categories_with_colors) / cmap.N))
    colors = (cmap.colors * ncyc)[: len(categories_with_colors)]
    base_colors = dict(zip(categories_with_colors, colors))
    if highlight_grp:
        base_colors[highlight_grp] = "#000000"
    return base_colors


def category_styledict(colordict, highlight_grp):
    """Generate a dictionary mapping categories to styles.

    The only styling implemented at present is converting the highlighted group to a dashed line.

    Args:
        colordict (dict) A mapping of categories to colors
        highlight_grp (string) The name of a group/category that should be highlighted

    Returns: (dict) A mapping of categories to matplotlib styles
    """
    return dict(
        (key, "--" if key == highlight_grp else "-") for key in colordict.keys()
    )


def _plot_lines(frame, x_col, y_col, ax, grp_col, colordict, cat_col, styledict, alpha):
    # plot the lines, one for each model group,
    # looking up the color by model type from above
    for grp_val in np.unique(frame[grp_col]):
        df = frame.loc[frame[grp_col] == grp_val]
        cat = df.iloc[0][cat_col]
        df.plot(
            x_col, y_col, ax=ax, c=colordict[cat], style=styledict[cat], legend=False, alpha=alpha
        )


def generate_plot_lines(colordict, label_fcn, styledict):
    plot_lines = []
    # plot_labs = []
    for cat_val in sorted(colordict.keys()):
        # http://matplotlib.org/users/legend_guide.html
        lin = mlines.Line2D(
            xdata=[],
            ydata=[],
            linestyle=styledict[cat_val],
            color=colordict[cat_val],
            label=label_fcn(cat_val),
        )
        plot_lines.append(lin)
        # plot_labs.append(mt)
    return plot_lines


def _config_axes(
    ax,
    x_ticks,
    y_ticks,
    x_lim,
    y_lim,
    title,
    title_fontsize,
    x_label,
    y_label,
    label_fontsize,
):
    if x_ticks is not None:
        ax.set_xticks(x_ticks)
    elif x_lim is not None:
        ax.set_xlim(x_lim)
    else:
        ax.set_xlim([0, 1.1])

    if y_ticks is not None:
        ax.set_yticks(y_ticks)
    elif y_lim is not None:
        ax.set_ylim(y_lim)
    else:
        ax.set_ylim([0, 1.1])
    ax.set_title(title, fontsize=title_fontsize)
    ax.set_ylabel(y_label, fontsize=label_fontsize)
    ax.set_xlabel(x_label, fontsize=label_fontsize)


def _get_leaf(path):
    return path.rsplit(".", 1)[-1]


def _no_op(arg):
    return arg


def plot_cats(
    frame,
    x_col,
    y_col,
    cat_col="model_type",
    grp_col="model_group_id",
    highlight_grp=None,
    title="",
    x_label="",
    y_label="",
    cmap_name="tab10",
    figsize=[12, 6],
    x_ticks=None,
    y_ticks=None,
    x_lim=None,
    y_lim=None,
    legend_loc=None,
    legend_fontsize=12,
    label_fontsize=12,
    title_fontsize=16,
    label_fcn=None,
    path_to_save=None,
    alpha=0.4,
    colordict=None,
    styledict=None,
):
    """Plot a line plot with each line colored by a category variable.

    Arguments:
        frame (DataFrame) -- a dataframe containing the data to be plotted
        x_col (string) -- name of the x-axis column
        y_col (string) -- name of the y-axis column
        cat_col (string) -- name of the catagory column to color lines
        grp_col (string) -- column that identifies each group of
                            (x_col, y_col) points for each line
        highlight_grp (string) -- name of group that should be highlighted in some way
            (like a baseline)
        title (string) -- allows specifying a custom title for the graph
        x_label (string) -- allows specifying a custom label for the x-axis
        y_label (string) -- allows specifying a custom label for the y-axis
        cmap_name (string) -- matplotlib color map name to use for plot
        figsize (tuple) -- figure size to pass to matplotlib
        x_ticks (sequence) -- optional ticks to use for x-axis
        y_ticks (sequence) -- optional ticks to use for y-axis
        x_lim (tuple) -- optional min-max to use for x-axis
        y_lim (tuple) -- optional min-max to use for y-axis
        legend_loc (string) -- allows specifying location of plot legend
        legend_fontsize (int) -- allows specifying font size for legend
        label_fontsize (int) -- allows specifying font size for axis labels
        title_fontsize (int) -- allows specifying font size for plot title
        label_fcn (method) -- function to map category names to more readable
                                names, accepting values of cat_col
        path_to_save (str) -- optional file path to save plot to disk
        alpha (float) -- value of alpha for plotting lines
        colordict (dict) -- optional dict mapping categories to colors
        styledict (dict) -- optional dict mapping categories to styles

    """
    fig, ax = plt.subplots(1, 1, figsize=figsize)

    # function for parsing cat_col values into more readable legend lables
    if label_fcn is None:
        label_fcn = _get_leaf if cat_col == "model_type" else _no_op

    categories = np.unique(frame[cat_col])

    colordict = colordict or category_colordict(cmap_name, categories, highlight_grp)
    styledict = styledict or category_styledict(colordict, highlight_grp)

    # plot the lines, one for each model group,
    # looking up the color by model type from above
    _plot_lines(frame, x_col, y_col, ax, grp_col, colordict, cat_col, styledict, alpha)

    # have to set the legend manually since we don't want one legend
    # entry per line on the plot, just one per model type.

    # I had to upgrade matplotlib to get handles working, otherwise
    # had to call like this with plot_labs as a separate list
    # plt.legend(plot_patches, plot_labs, loc=4, fontsize=10)
    plot_lines = generate_plot_lines(colordict, label_fcn, styledict)

    plt.legend(handles=plot_lines, loc=legend_loc, fontsize=legend_fontsize)

    _config_axes(
        ax=ax,
        x_ticks=x_ticks,
        y_ticks=y_ticks,
        x_lim=x_lim,
        y_lim=y_lim,
        title=title,
        title_fontsize=title_fontsize,
        x_label=x_label,
        y_label=y_label,
        label_fontsize=label_fontsize,
    )

    plt.show()

    if path_to_save:
        plt.savefig(path_to_save)



import datetime
import functools
import importlib
import random
import tempfile
from contextlib import contextmanager
from unittest import mock

import matplotlib
import numpy as np
import pandas as pd
import testing.postgresql
from descriptors import cachedproperty
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from triage.component.catwalk.db import ensure_db
from triage.component.catwalk.storage import MatrixStore, ProjectStorage
from triage.component.catwalk.utils import filename_friendly_hash
from triage.component.results_schema import Model, Matrix
from triage.experiments import CONFIG_VERSION
from triage.util.structs import FeatureNameList

from tests.results_tests.factories import init_engine, session, MatrixFactory

matplotlib.use("Agg")

from matplotlib import pyplot as plt  # noqa


def fake_labels(length):
    return np.array([random.choice([True, False]) for i in range(0, length)])


class MockTrainedModel:
    def predict_proba(self, dataset):
        return np.random.rand(len(dataset), len(dataset))


class MockMatrixStore(MatrixStore):
    def __init__(
        self,
        matrix_type,
        matrix_uuid,
        label_count,
        db_engine,
        init_labels=None,
        metadata_overrides=None,
        matrix=None,
        init_as_of_dates=None,
    ):
        base_metadata = {
            "feature_start_time": datetime.date(2014, 1, 1),
            "end_time": datetime.date(2015, 1, 1),
            "as_of_date_frequency": "1y",
            "matrix_id": "some_matrix",
            "label_name": "label",
            "label_timespan": "3month",
            "indices": MatrixStore.indices,
            "matrix_type": matrix_type,
            "as_of_times": [datetime.date(2014, 10, 1), datetime.date(2014, 7, 1)],
        }
        metadata_overrides = metadata_overrides or {}
        base_metadata.update(metadata_overrides)
        if matrix is None:
            matrix = pd.DataFrame.from_dict(
                {
                    "entity_id": [1, 2],
                    "as_of_date": [pd.Timestamp(2014, 10, 1), pd.Timestamp(2014, 7, 1)],
                    "feature_one": [3, 4],
                    "feature_two": [5, 6],
                    "label": [7, 8],
                }
            ).set_index(MatrixStore.indices)
        if init_labels is None:
            init_labels = []
        labels = matrix.pop("label")
        self.matrix_label_tuple = matrix, labels
        self.metadata = base_metadata
        self.label_count = label_count
        self.init_labels = pd.Series(init_labels, dtype='float64')
        self.matrix_uuid = matrix_uuid
        self.init_as_of_dates = init_as_of_dates or []

        session = sessionmaker(db_engine)()
        session.add(Matrix(matrix_uuid=matrix_uuid))
        session.commit()

    @property
    def as_of_dates(self):
        """The list of as-of-dates in the matrix"""
        return self.init_as_of_dates or self.metadata["as_of_times"]

    @property
    def labels(self):
        if len(self.init_labels) > 0:
            return self.init_labels
        else:
            return fake_labels(self.label_count)


def fake_trained_model(
    db_engine,
    train_matrix_uuid="efgh",
    train_end_time=datetime.datetime(2016, 1, 1)
):
    """Creates and stores a trivial trained model and training matrix

    Args:
        db_engine (sqlalchemy.engine)

    Returns:
        (int) model id for database retrieval
    """
    session = sessionmaker(db_engine)()
    session.merge(Matrix(matrix_uuid=train_matrix_uuid))

    # Create the fake trained model and store in db
    trained_model = MockTrainedModel()
    db_model = Model(
        model_hash="abcd",
        train_matrix_uuid=train_matrix_uuid,
        train_end_time=train_end_time,
    )
    session.add(db_model)
    session.commit()
    model_id = db_model.model_id
    session.close()
    return trained_model, model_id


def matrix_metadata_creator(**override_kwargs):
    """Create a sample valid matrix metadata with optional overrides

    Args:
    **override_kwargs: Keys and values to override in the metadata

    Returns: (dict)
    """
    base_metadata = {
        "feature_start_time": datetime.date(2012, 12, 20),
        "end_time": datetime.date(2016, 12, 20),
        "label_name": "label",
        "as_of_date_frequency": "1w",
        "max_training_history": "5y",
        "matrix_id": "tester-1",
        "state": "active",
        "cohort_name": "default",
        "label_timespan": "1y",
        "metta-uuid": "1234",
        "matrix_type": "test",
        "feature_names": FeatureNameList(["ft1", "ft2"]),
        "feature_groups": ["all: True"],
        "indices": MatrixStore.indices,
        "as_of_times": [datetime.date(2016, 12, 20)],
    }
    for override_key, override_value in override_kwargs.items():
        base_metadata[override_key] = override_value

    return base_metadata


def matrix_creator():
    """Return a sample matrix."""

    source_dict = {
        "entity_id": [1, 2],
        "as_of_date": [pd.Timestamp(2016, 1, 1), pd.Timestamp(2016, 1, 1)],
        "feature_one": [3, 4],
        "feature_two": [5, 6],
        "label": [0, 1],
    }
    return pd.DataFrame.from_dict(source_dict)


def get_matrix_store(project_storage, matrix=None, metadata=None, write_to_db=True):
    """Return a matrix store associated with the given project storage.
    Also adds an entry in the matrices table if it doesn't exist already

    Args:
        project_storage (triage.component.catwalk.storage.ProjectStorage) A project's storage
        matrix (dataframe, optional): A matrix to store. Defaults to the output of matrix_creator()
        metadata (dict, optional): matrix metadata.
            defaults to the output of matrix_metadata_creator()
    """
    if matrix is None:
        matrix = matrix_creator()
    if not metadata:
        metadata = matrix_metadata_creator()
    matrix["as_of_date"] = matrix["as_of_date"].apply(pd.Timestamp)
    matrix.set_index(MatrixStore.indices, inplace=True)
    matrix_store = (project_storage.matrix_storage_engine()
                    .get_store(filename_friendly_hash(metadata)))
    matrix_store.metadata = metadata
    new_matrix = matrix.copy()
    labels = new_matrix.pop(matrix_store.label_column_name)
    matrix_store.matrix_label_tuple = new_matrix, labels
    matrix_store.save()
    matrix_store.clear_cache()
    if write_to_db:
        if (
            session.query(Matrix).filter(
                Matrix.matrix_uuid == matrix_store.uuid).count()
            == 0
        ):
            MatrixFactory(matrix_uuid=matrix_store.uuid)
            session.commit()
    return matrix_store


@contextmanager
def rig_engines():
    """Set up a db engine and project storage engine

    Yields (tuple) (database engine, project storage engine)
    """
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        init_engine(db_engine)
        with tempfile.TemporaryDirectory() as temp_dir:
            project_storage = ProjectStorage(temp_dir)
            yield db_engine, project_storage


def populate_source_data(db_engine):
    complaints = [
        (1, "2010-10-01", 5),
        (1, "2011-10-01", 4),
        (1, "2011-11-01", 4),
        (1, "2011-12-01", 4),
        (1, "2012-02-01", 5),
        (1, "2012-10-01", 4),
        (1, "2013-10-01", 5),
        (2, "2010-10-01", 5),
        (2, "2011-10-01", 5),
        (2, "2011-11-01", 4),
        (2, "2011-12-01", 4),
        (2, "2012-02-01", 6),
        (2, "2012-10-01", 5),
        (2, "2013-10-01", 6),
        (3, "2010-10-01", 5),
        (3, "2011-10-01", 3),
        (3, "2011-11-01", 4),
        (3, "2011-12-01", 4),
        (3, "2012-02-01", 4),
        (3, "2012-10-01", 3),
        (3, "2013-10-01", 4),
    ]

    entity_zip_codes = [(1, "60120"), (2, "60123"), (3, "60123")]

    zip_code_demographics = [
        ("60120", "hispanic", "2011-01-01"),
        ("60123", "white", "2011-01-01"),
    ]

    zip_code_events = [("60120", "2012-10-01", 1), ("60123", "2012-10-01", 10)]

    events = [
        (1, 1, "2011-01-01"),
        (1, 1, "2011-06-01"),
        (1, 1, "2011-09-01"),
        (1, 1, "2012-01-01"),
        (1, 1, "2012-01-10"),
        (1, 1, "2012-06-01"),
        (1, 1, "2013-01-01"),
        (1, 0, "2014-01-01"),
        (1, 1, "2015-01-01"),
        (2, 1, "2011-01-01"),
        (2, 1, "2011-06-01"),
        (2, 1, "2011-09-01"),
        (2, 1, "2012-01-01"),
        (2, 1, "2013-01-01"),
        (2, 1, "2014-01-01"),
        (2, 1, "2015-01-01"),
        (3, 0, "2011-01-01"),
        (3, 0, "2011-06-01"),
        (3, 0, "2011-09-01"),
        (3, 0, "2012-01-01"),
        (3, 0, "2013-01-01"),
        (3, 1, "2014-01-01"),
        (3, 0, "2015-01-01"),
    ]

    db_engine.execute(
        """create table cat_complaints (
        entity_id int,
        as_of_date date,
        cat_sightings int
        )"""
    )

    db_engine.execute(
        """create table entity_zip_codes (
        entity_id int,
        zip_code text
        )"""
    )

    db_engine.execute(
        "create table zip_code_demographics (zip_code text, ethnicity text, as_of_date date)")
    for demographic_row in zip_code_demographics:
        db_engine.execute(
            "insert into zip_code_demographics values (%s, %s, %s)", demographic_row)

    for entity_zip_code in entity_zip_codes:
        db_engine.execute(
            "insert into entity_zip_codes values (%s, %s)", entity_zip_code
        )

    db_engine.execute(
        """create table zip_code_events (
        zip_code text,
        as_of_date date,
        num_events int
    )"""
    )
    for zip_code_event in zip_code_events:
        db_engine.execute(
            "insert into zip_code_events values (%s, %s, %s)", zip_code_event
        )

    for complaint in complaints:
        db_engine.execute(
            "insert into cat_complaints values (%s, %s, %s)", complaint)

    db_engine.execute(
        """create table events (
        entity_id int,
        outcome int,
        outcome_date date
    )"""
    )

    for event in events:
        db_engine.execute("insert into events values (%s, %s, %s)", event)


def sample_config():
    temporal_config = {
        "feature_start_time": "2010-01-01",
        "feature_end_time": "2014-01-01",
        "label_start_time": "2011-01-01",
        "label_end_time": "2014-01-01",
        "model_update_frequency": "1year",
        "training_label_timespans": ["6months"],
        "test_label_timespans": ["6months"],
        "training_as_of_date_frequencies": "1day",
        "test_as_of_date_frequencies": "3months",
        "max_training_histories": ["6months"],
        "test_durations": ["1months"],
    }

    scoring_config = {
        "testing_metric_groups": [
            {"metrics": ["precision@"], "thresholds": {"top_n": [2]}}
        ],
        "training_metric_groups": [
            {"metrics": ["precision@"], "thresholds": {"top_n": [3]}}
        ],
        "subsets": [
            {
                "name": "evens",
                "query": """\
                    select distinct entity_id
                    from events
                    where entity_id %% 2 = 0
                    and outcome_date < '{as_of_date}'::date
                """,
            },
        ]
    }

    grid_config = {
        "sklearn.tree.DecisionTreeClassifier": {
            "min_samples_split": [10, 100],
            "max_depth": [3, 5],
            "criterion": ["gini"],
        }
    }

    feature_config = [
        {
            "prefix": "entity_features",
            "from_obj": "cat_complaints",
            "knowledge_date_column": "as_of_date",
            "aggregates_imputation": {"all": {"type": "constant", "value": 0}},
            "aggregates": [{"quantity": "cat_sightings", "metrics": ["count", "avg"]}],
            "intervals": ["1year"],
            "groups": ["entity_id"],
        },
        {
            "prefix": "zip_code_features",
            "from_obj": "entity_zip_codes join zip_code_events using (zip_code)",
            "knowledge_date_column": "as_of_date",
            "aggregates_imputation": {"all": {"type": "constant", "value": 0}},
            "aggregates": [{"quantity": "num_events", "metrics": ["max", "min"]}],
            "intervals": ["1year"],
            "groups": ["entity_id", "zip_code"],
        },
    ]

    cohort_config = {
        "query": "select distinct(entity_id) from events "
                 "where '{as_of_date}'::date >= outcome_date",
        "name": "has_past_events",
    }

    label_config = {
        "query": """
            select
            events.entity_id,
            bool_or(outcome::bool)::integer as outcome
            from events
            where '{as_of_date}'::date <= outcome_date
                and outcome_date < '{as_of_date}'::date + interval '{label_timespan}'
                group by entity_id
        """,
        "name": "custom_label_name",
        "include_missing_labels_in_train_as": False,
    }
    bias_audit_config = {
        'from_obj_query': 'select * from zip_code_demographics join entity_zip_codes using (zip_code)',
        'attribute_columns': ['ethnicity'],
        'knowledge_date_column': 'as_of_date',
        'entity_id_column': 'entity_id',
        'ref_groups_method': 'predefined',
        'ref_groups': {'ethnicity': 'white'},
        'thresholds': {
            'percentiles': [],
            'top_n': [2]
        }
    }

    return {
        "config_version": CONFIG_VERSION,
        "random_seed": 1234,
        "label_config": label_config,
        "entity_column_name": "entity_id",
        "model_comment": "test2-final-final",
        "model_group_keys": ["label_name", "label_type", "custom_key"],
        "feature_aggregations": feature_config,
        "cohort_config": cohort_config,
        "temporal_config": temporal_config,
        "grid_config": grid_config,
        "bias_audit_config": bias_audit_config,
        "prediction": {"rank_tiebreaker": "random"},
        "scoring": scoring_config,
        "user_metadata": {"custom_key": "custom_value"},
        "individual_importance": {"n_ranks": 2},
    }


@contextmanager
def assert_plot_figures_added():
    num_figures_before = plt.gcf().number
    yield
    num_figures_after = plt.gcf().number
    assert num_figures_before < num_figures_after


class CallSpy:
    """Callable-wrapper and -patcher to record invocations.

    ``CallSpy``, (unlike ``Mock``), makes it easy to wrap callables for
    the express purpose of recording how they're invoked – without
    modifying functionality. And ``CallSpy``, (unlike ``Mock``),
    reproduces the descriptor interface, such that methods can be
    patched and proxied for this purpose, as easily as functions.

    For example, as a context manager::

        with CallSpy('my_module.MyClass.my_method') as spy:
            ...

        assert (('arg0',), {'param0': 0}) in spy.calls

    """

    def __init__(self, signature):
        self.calls = []
        self.signature = signature

    @cachedproperty
    def target_path(self):
        return self.signature.split('.')

    @cachedproperty
    def target_name(self):
        return self.target_path[-1]

    @cachedproperty
    def target_base(self):
        # walk target path until can no longer import it as a module path
        for index in range(len(self.target_path)):
            path_parts = self.target_path[:(index + 1)]
            import_path = '.'.join(path_parts)

            try:
                base = importlib.import_module(import_path)
            except ImportError:
                # we've imported all that we can import
                # walk the remainder by attribute access
                remainder = self.target_path[index:-1]
                for part in remainder:
                    base = getattr(base, part)

                return base

        raise ValueError(f"cannot patch signature {self.signature!r}")

    @cachedproperty
    def target_object(self):
        return getattr(self.target_base, self.target_name)

    @cachedproperty
    def patch(self):
        return mock.patch.object(self.target_base, self.target_name, new=self)

    def start(self):
        if not callable(self.target_object):
            # 1. ensure target_object set before patching
            # 2. check that it's sane (needn't be done here but reasonable)
            raise TypeError(
                f"signature target not callable {self.target_object!r}")

        self.patch.start()

    def stop(self):
        self.patch.stop()

    def __call__(self, *args, **kwargs):
        self.calls.append((args, kwargs))
        return self.target_object(*args, **kwargs)

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        return functools.partial(self, instance)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc):
        self.stop()



from functools import partial
import pandas as pd
import numpy as np

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

def downcast_matrix(df):
    """Downcast the numeric values of a matrix.

    This will make the matrix use less memory by turning, for instance,
    int64 columns into int32 columns.

    First converts floats and then integers.

    Operates on the dataframe as passed, without doing anything to the index.
    Callers may pass an index-less dataframe if they wish to re-add the index afterwards
    and save memory on the index storage.
    """
    logger.spam("Downcasting matrix.")
    logger.spam(f"Starting memory usage: {df.memory_usage(deep=True).sum()} bytes")
    logger.spam(f"Initial types: \n {df.dtypes}")
    new_df = df.apply(lambda x: x.astype(np.float32))

    logger.spam("Downcasting matrix completed.")
    logger.spam(f"Final memory usage: {new_df.memory_usage(deep=True).sum()} bytes")
    logger.spam(f"Final data types: \n {new_df.dtypes}")

    return new_df



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import os
import pandas as pd
import numpy as np

from .plotting import plot_cats, category_colordict, category_styledict
from .utils import str_in_sql


class ModelGroupPerformancePlotter:
    def __init__(self, distance_from_best_table, directory=None):
        """Generate a plot illustrating the performance of a model group over time

        Args:
            distance_from_best_table (audition.DistanceFromBestTable)
                A pre-populated distance-from-best database table
        """
        self.distance_from_best_table = distance_from_best_table
        self.directory = directory
        self.colordict = None
        self.styledict = None
        self.highlight_grp = "best case"
        self.cmap_name = "tab10"

    def plot_all(self, metric_filters, model_group_ids, train_end_times):
        """For each metric, plot the value of that metric over time

        Arguments:
            metric_filters (list) The metrics to plot. Each element should be
                a dict with the following keys:

                metric (string) -- model evaluation metric, such as 'precision@'
                parameter (string) -- model evaluation metric parameter,
                    such as '300_abs'
            model_group_ids (list) - Model group ids to include in the plot
            train_end_times (list) - Train end times to include in the plot

        """
        for metric_filter in metric_filters:
            logger.debug(
                f"Plotting model group performance for {metric_filter}, {train_end_times}",
            )
            df = self.generate_plot_data(
                metric=metric_filter["metric"],
                parameter=metric_filter["parameter"],
                model_group_ids=model_group_ids,
                train_end_times=train_end_times,
            )

            # set stable colors/styles by model type
            categories = np.unique(df['model_type'])
            if not self.colordict:
                self.colordict = category_colordict(self.cmap_name, categories, self.highlight_grp)
            if not self.styledict:
                self.styledict = category_styledict(self.colordict, self.highlight_grp)

            self.plot(
                metric=metric_filter["metric"],
                parameter=metric_filter["parameter"],
                df_metric=df,
                train_end_times=train_end_times,
                directory=self.directory,
            )

    def generate_plot_data(self, metric, parameter, model_group_ids, train_end_times):
        """Fetch data necessary for producing the plot from the distance table

        Arguments:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter,
                such as '300_abs'
            model_group_ids (list) - Model group ids to include in the dataset
            train_end_times (list) - Train end times to include in the dataset

        Returns: (pandas.DataFrame) The relevant models and their performance
        on the given metric over time
        """

        df = pd.read_sql(
            """
            select distinct on(model_group_id, metric, parameter, train_end_time, raw_value, model_type) * from (
                select
                    model_group_id,
                    metric,
                    parameter,
                    train_end_time,
                    raw_value,
                    mg.model_type as model_type
                from {dist_table} as dist
                join triage_metadata.model_groups mg using (model_group_id)
                where model_group_id in ({model_group_ids})
                union
                select
                    0 as model_group_id,
                    metric,
                    parameter,
                    train_end_time,
                    best_case as raw_value,
                    'best case' as model_type
                from {dist_table}
            ) as t
            where metric || parameter = '{metric}{parameter}'
            and train_end_time in ({train_end_times})
            order by model_group_id asc, train_end_time asc
            """.format(
                metric=metric,
                parameter=parameter,
                dist_table=self.distance_from_best_table.distance_table,
                model_group_ids=str_in_sql(model_group_ids),
                train_end_times=str_in_sql(train_end_times)
            ),
            self.distance_from_best_table.db_engine,
        )

        return df

    def plot(
        self,
        metric,
        parameter,
        df_metric,
        train_end_times,
        directory,
        **plt_format_args,
    ):
        """Draw the plot representing the given data

        Arguments:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter, such as '300_abs'
            df_metric (pandas.DataFrame)
            train_end_times (list) - Train end times to use for ticks
            **plt_format_args -- formatting arguments passed through to plot_cats()
        """
        cat_col = "model_type"
        plt_title = "{} {} over time".format(metric, parameter)

        # when setting the ticks, matplotlib sometimes has problems with datetimes given
        # as np.datetime64 objects, and converting from them to datetimes is ugly.
        # to get around this, we use the train_end_times given to the plot call as ticks
        # But to be defensive, we verify that these two versions of the list are the same
        for given_time, matrix_time in zip(
            train_end_times, sorted(df_metric["train_end_time"].unique())
        ):
            given_time_as_numpy = np.datetime64(given_time)
            if given_time_as_numpy != matrix_time:
                raise ValueError(
                    "Train times given to the plotter do not match up with those "
                    "extracted from the database: "
                    "{} (given time) does not equal {} (matrix time)".format(
                        given_time_as_numpy, matrix_time
                    )
                )
        if directory:
            path_to_save = os.path.join(
                directory, f"metric_over_time_{metric}{parameter}.png"
            )
        else:
            path_to_save = None

        plot_cats(
            frame=df_metric,
            x_col="train_end_time",
            y_col="raw_value",
            cat_col=cat_col,
            highlight_grp=self.highlight_grp,
            title=plt_title,
            x_label="train end time",
            y_label="value of {}".format(metric),
            x_ticks=train_end_times,
            path_to_save=path_to_save,
            colordict=self.colordict,
            styledict=self.styledict,
            **plt_format_args,
        )



from unittest.mock import patch
from unittest import TestCase
import yaml
import matplotlib

matplotlib.use("Agg")
from triage.component.timechop import Timechop # noqa
from triage.component.timechop.plotting import visualize_chops # noqa


class VisualizeChopTest(TestCase):
    @property
    def chopper(self):
        # create a valid Timechop chopper
        # least brittle current way of doing this is by loading the
        # example_experiment_config.yaml file, because that is a
        # diligently updated file. If Timechop config changes, the
        # example config should change too
        with open("example/config/experiment.yaml") as fd:
            experiment_config = yaml.full_load(fd)
        return Timechop(**(experiment_config["temporal_config"]))

    # hard to make many assertions, but we can make sure it gets to the end
    # and shows the contents.

    # we do one such test case to work out each combination of boolean arguments
    def test_default_args(self):
        with patch("triage.component.timechop.plotting.plt.show") as show_patch:
            visualize_chops(self.chopper)
            assert show_patch.called

    def test_no_as_of_times(self):
        with patch("triage.component.timechop.plotting.plt.show") as show_patch:
            visualize_chops(self.chopper, show_as_of_times=False)
            assert show_patch.called

    def test_no_boundaries(self):
        with patch("triage.component.timechop.plotting.plt.show") as show_patch:
            visualize_chops(self.chopper, show_boundaries=False)
            assert show_patch.called

    def test_no_boundaries_or_as_of_times(self):
        with patch("triage.component.timechop.plotting.plt.show") as show_patch:
            visualize_chops(self.chopper, show_as_of_times=False, show_boundaries=False)
            assert show_patch.called



"""Split results into model_metadata, test_results, and train_resultss

Revision ID: 89a8ce240bae
Revises: 7d57d1cf3429
Create Date: 2018-03-27 15:12:45.249609

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "89a8ce240bae"
down_revision = "7d57d1cf3429"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("CREATE SCHEMA IF NOT EXISTS model_metadata")
    op.execute("CREATE SCHEMA IF NOT EXISTS test_results")
    op.execute("CREATE SCHEMA IF NOT EXISTS train_results")
    op.execute(
        "ALTER TABLE results.models SET SCHEMA model_metadata;"
        + " ALTER TABLE results.model_groups SET SCHEMA model_metadata;"
        + " ALTER TABLE results.experiments SET SCHEMA model_metadata;"
        + " ALTER TABLE results.list_predictions SET SCHEMA model_metadata;"
        + " ALTER TABLE results.predictions RENAME TO test_predictions;"
        + " ALTER TABLE results.test_predictions SET SCHEMA test_results;"
        + " ALTER TABLE results.evaluations RENAME TO test_evaluations;"
        + " ALTER TABLE results.test_evaluations SET SCHEMA test_results;"
        + " ALTER TABLE results.individual_importances SET SCHEMA test_results;"
        + " ALTER TABLE results.feature_importances SET SCHEMA train_results;"
    )

    op.execute("ALTER TABLE model_metadata.models ADD COLUMN model_size real")
    op.create_table(
        "matrices",
        sa.Column("matrix_id", sa.String(), nullable=True),
        sa.Column("matrix_uuid", sa.String(), nullable=False),
        sa.Column("matrix_type", sa.String(), nullable=True),
        sa.Column("labeling_window", sa.Interval(), nullable=True),
        sa.Column("n_examples", sa.Integer(), nullable=True),
        sa.Column(
            "creation_time",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("lookback_duration", sa.Interval(), nullable=True),
        sa.Column("feature_start_time", sa.DateTime(), nullable=True),
        sa.Column(
            "matrix_metadata", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.PrimaryKeyConstraint("matrix_uuid"),
        schema="model_metadata",
    )
    op.create_index(
        op.f("ix_model_metadata_matrices_matrix_uuid"),
        "matrices",
        ["matrix_uuid"],
        unique=True,
        schema="model_metadata",
    )

    # Create train_predictions and train_evaluations
    op.create_table(
        "train_evaluations",
        sa.Column("model_id", sa.Integer(), nullable=False),
        sa.Column("evaluation_start_time", sa.DateTime(), nullable=False),
        sa.Column("evaluation_end_time", sa.DateTime(), nullable=False),
        sa.Column("as_of_date_frequency", sa.Interval(), nullable=False),
        sa.Column("metric", sa.String(), nullable=False),
        sa.Column("parameter", sa.String(), nullable=False),
        sa.Column("value", sa.Numeric(), nullable=True),
        sa.Column("num_labeled_examples", sa.Integer(), nullable=True),
        sa.Column("num_labeled_above_threshold", sa.Integer(), nullable=True),
        sa.Column("num_positive_labels", sa.Integer(), nullable=True),
        sa.Column("sort_seed", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["model_id"], ["model_metadata.models.model_id"]),
        sa.PrimaryKeyConstraint(
            "model_id",
            "evaluation_start_time",
            "evaluation_end_time",
            "as_of_date_frequency",
            "metric",
            "parameter",
        ),
        schema="train_results",
    )
    op.create_table(
        "train_predictions",
        sa.Column("model_id", sa.Integer(), nullable=False),
        sa.Column("entity_id", sa.BigInteger(), nullable=False),
        sa.Column("as_of_date", sa.DateTime(), nullable=False),
        sa.Column("score", sa.Numeric(), nullable=True),
        sa.Column("label_value", sa.Integer(), nullable=True),
        sa.Column("rank_abs", sa.Integer(), nullable=True),
        sa.Column("rank_pct", sa.Float(), nullable=True),
        sa.Column("matrix_uuid", sa.Text(), nullable=True),
        sa.Column("test_label_timespan", sa.Interval(), nullable=True),
        sa.ForeignKeyConstraint(
            ["matrix_uuid"], ["model_metadata.matrices.matrix_uuid"]
        ),
        sa.ForeignKeyConstraint(["model_id"], ["model_metadata.models.model_id"]),
        sa.PrimaryKeyConstraint("model_id", "entity_id", "as_of_date"),
        schema="train_results",
    )

    # Foreign Keys for the models, test_predictions, and train_predictions tables
    op.execute(
        """ALTER TABLE model_metadata.models
                  ADD CONSTRAINT matrix_uuid_for_models
                  FOREIGN KEY (train_matrix_uuid)
                  REFERENCES model_metadata.matrices(matrix_uuid); """
    )

    op.execute(
        """ALTER TABLE test_results.test_predictions
                  ADD CONSTRAINT matrix_uuid_for_testpred
                  FOREIGN KEY (matrix_uuid)
                  REFERENCES model_metadata.matrices(matrix_uuid); """
    )

    op.execute(
        """ALTER TABLE train_results.train_predictions
                  ADD CONSTRAINT matrix_uuid_for_trainpred
                  FOREIGN KEY (matrix_uuid)
                  REFERENCES model_metadata.matrices(matrix_uuid); """
    )

    op.execute("DROP SCHEMA IF EXISTS results")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("CREATE SCHEMA IF NOT EXISTS results")

    op.execute(
        """ALTER TABLE model_metadata.models DROP CONSTRAINT "matrix_uuid_for_models";
        ALTER TABLE test_results.test_predictions DROP CONSTRAINT "matrix_uuid_for_testpred";
        ALTER TABLE train_results.train_predictions DROP CONSTRAINT "matrix_uuid_for_trainpred"; """
    )

    op.execute(
        "ALTER TABLE model_metadata.models SET SCHEMA results;"
        + " ALTER TABLE model_metadata.model_groups SET SCHEMA results;"
        + " ALTER TABLE model_metadata.experiments SET SCHEMA results;"
        + " ALTER TABLE model_metadata.list_predictions SET SCHEMA results;"
        + " ALTER TABLE test_results.test_predictions RENAME TO predictions;"
        + " ALTER TABLE test_results.predictions SET SCHEMA results;"
        + " ALTER TABLE test_results.test_evaluations RENAME TO evaluations;"
        + " ALTER TABLE test_results.evaluations SET SCHEMA results;"
        + " ALTER TABLE test_results.individual_importances SET SCHEMA results;"
        + " ALTER TABLE train_results.feature_importances SET SCHEMA results;"
    )

    op.drop_index(
        op.f("ix_model_metadata_matrices_matrix_uuid"),
        table_name="matrices",
        schema="model_metadata",
    )
    op.execute("ALTER TABLE results.models DROP COLUMN IF EXISTS model_size")
    op.execute("DROP TABLE model_metadata.matrices CASCADE")
    op.execute("DROP TABLE train_results.train_predictions CASCADE")
    op.execute("DROP TABLE train_results.train_evaluations CASCADE")

    op.execute("DROP SCHEMA IF EXISTS model_metadata")
    op.execute("DROP SCHEMA IF EXISTS train_results")
    op.execute("DROP SCHEMA IF EXISTS test_results")
    # ### end Alembic commands ###



"""Change importance to float

Revision ID: 72ac5cbdca05
Revises: 264245ddfce2
Create Date: 2017-09-01 14:31:09.302828

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "72ac5cbdca05"
down_revision = "264245ddfce2"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        table_name="individual_importances",
        column_name="importance_score",
        type_=sa.Float(),
        schema="results",
        postgresql_using="importance_score::double precision",
    )


def downgrade():
    op.alter_column(
        table_name="individual_importances",
        column_name="importance_score",
        type_=sa.Text(),
        schema="results",
    )
